from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Annalist form building utilities
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"




"""
Namespace vocabulary functions to support testing
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2015, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os

from django.core.urlresolvers   import resolve, reverse

from annalist.util              import valid_id
from annalist.identifiers       import RDF, RDFS, ANNAL
from annalist                   import layout

from .entity_testfielddesc      import get_field_description, get_bound_field
from .entity_testentitydata     import entitydata_list_type_url
from .entity_testutils import (
    collection_dir, 
    collection_entity_view_url,
    context_field_row
    )
from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )

#   -----------------------------------------------------------------------------
#
#   Directory generating functions
#
#   -----------------------------------------------------------------------------

def recordvocab_dir(coll_id="testcoll", vocab_id="testvocab"):
    return collection_dir(coll_id) + layout.COLL_VOCAB_PATH%{'id': vocab_id} + "/"

#   -----------------------------------------------------------------------------
#
#   URI generating functions
#
#   -----------------------------------------------------------------------------

def recordvocab_coll_url(site, coll_id="testcoll", vocab_id="testvocab"):
    return (
        site._entityurl + 
        layout.SITE_COLL_PATH%{'id': coll_id} + "/" + 
        layout.COLL_VOCAB_PATH%{'id': vocab_id} + "/"
        )

def recordvocab_url(coll_id, vocab_id):
    """
    URI for namespace vocabulary description data; also view using default entity view
    """
    if not valid_id(vocab_id):
        return None
        # vocab_id = "___"
    return collection_entity_view_url(coll_id=coll_id, type_id="_vocab", entity_id=vocab_id)

def recordvocab_edit_url(action=None, coll_id=None, vocab_id=None):
    """
    URI for vocabulary description editing view
    """
    viewname = ( 
        'AnnalistEntityDataView'        if action == "view"   else
        'AnnalistEntityNewView'         if action == "new"    else
        'AnnalistEntityEditView'        if action == "copy"   else
        'AnnalistEntityEditView'        if action == "edit"   else
        'AnnalistRecordVocabDeleteView' if action == "delete" else
        'unknown'
        )
    kwargs = {'coll_id': coll_id, 'type_id': "_vocab", 'view_id': "Vocab_view"}
    if action != "delete":
        kwargs.update({'action': action})
    if vocab_id:
        if valid_id(vocab_id):
            kwargs.update({'entity_id': vocab_id})
        else:
            kwargs.update({'entity_id': "___"})
    return reverse(viewname, kwargs=kwargs)

#   -----------------------------------------------------------------------------
#
#   ----- RecordVocab data
#
#   -----------------------------------------------------------------------------

def recordvocab_init_keys():
    keys = set(
        [ 'annal:id'
        , 'annal:type_id'
        , 'annal:type'
        , 'annal:url'
        , 'rdfs:label'
        , 'rdfs:comment'
        , 'annal:uri'
        , 'rdfs:seeAlso'
        ])
    return keys

def recordvocab_value_keys():
    return (recordvocab_init_keys())

def recordvocab_load_keys():
    return recordvocab_value_keys() | {'@id', '@type', '@context'}

def recordvocab_create_values(
        coll_id="testcoll", vocab_id="testvocab", 
        vocab_uri="test:testvocab#",
        update="RecordVocab"):
    """
    Entity values used when creating a namespace vocabulary entity
    """
    d = (
        { 'rdfs:label':                 "%s %s/_vocab/%s"%(update, coll_id, vocab_id)
        , 'rdfs:comment':               "%s help for %s in collection %s"%(update, vocab_id, coll_id)
        , 'rdfs:seeAlso':               []
        })
    if vocab_uri:
        d['annal:uri'] = vocab_uri
    return d

def recordvocab_values(
        coll_id="testcoll", vocab_id="testvocab", 
        vocab_uri="test:testvocab#",
        update="RecordVocab"):
    d = recordvocab_create_values(
        coll_id, vocab_id, 
        vocab_uri=vocab_uri,
        update=update
        ).copy()
    d.update(
        { 'annal:id':                   vocab_id
        , 'annal:type_id':              "_vocab"
        , 'annal:type':                 "annal:Vocabulary"
        , 'annal:url':                  recordvocab_url(coll_id, vocab_id)
        })
    return d

def recordvocab_read_values(
        coll_id="testcoll", vocab_id="testvocab",
        vocab_uri="test:testvocab#",
        update="RecordVocab"):
    d = recordvocab_values(coll_id, vocab_id, vocab_uri=vocab_uri, update=update)
    d.update(
        { '@id':            "_vocab/%s"%(vocab_id)
        , '@type':          ["annal:Vocabulary"]
        , '@context':       [{"@base": "../../"}, "../../coll_context.jsonld"]
        })
    return d


#   -----------------------------------------------------------------------------
#
#   ----- Entity data in vocabulary namespace view
#
#   -----------------------------------------------------------------------------

def vocab_view_context_data(
        coll_id="testcoll", vocab_id="", orig_id=None, action=None, 
        vocab_label=None,
        vocab_descr=None,
        vocab_uri=None,
        vocab_seealso=[],
        update="RecordVocab",
        continuation_url=None
    ):
    if vocab_id:
        def_vocab_label = "%s %s/_vocab/%s"%(update, coll_id, vocab_id)
    else:
        def_vocab_label = "%s data ... (%s/_vocab/)"%(update, coll_id)
    if vocab_label is None:
        vocab_label = def_vocab_label
    if vocab_uri is None:
       vocab_uri = recordvocab_url(coll_id=coll_id, vocab_id=vocab_id)
    if continuation_url is None:
        continuation_url = entitydata_list_type_url(coll_id, "_vocab")
    view_label = "Vocabulary namespace"
    view_title = (
        "%s - %s - Collection %s"%(vocab_label, view_label, coll_id) if vocab_label
        else
        "%s - Collection %s"%(view_label, coll_id)
        )
    context_dict = (
        { 'title':              view_title
        , 'heading':            view_label
        , 'coll_id':            coll_id
        , 'type_id':            '_vocab'
        , 'view_id':            'Vocab_view'
        , 'entity_id':          vocab_id or ""
        , 'orig_id':            orig_id
        , 'orig_type':          layout.VOCAB_TYPEID
        , 'record_type':        "annal:Vocabulary"
        , 'continuation_url':   continuation_url
        , 'fields':
          [ context_field_row(
              get_bound_field("Vocab_id",          vocab_id),               # 0 (0,0)
              )
          , context_field_row(
              get_bound_field("Entity_label",      vocab_label)             # 1 (1,0)
              )
          , context_field_row(
              get_bound_field("Entity_comment",    vocab_descr)             # 2 (2,0)
              )
          , context_field_row(
              get_bound_field("Vocab_uri",         vocab_uri)               # 3 (3,0)
              )
          , get_bound_field("Entity_see_also_r",   vocab_seealso)           # 4 (4, 0)
          ]
        })
    if orig_id:
        context_dict['orig_id']     = orig_id
    if action:  
        context_dict['action']      = action
    return context_dict

def vocab_view_form_data(
        coll_id="testcoll", orig_coll=None,
        vocab_id="", orig_id=None, 
        action=None, cancel=None, close=None, edit=None, copy=None, task=None,
        update="RecordVocab",
        vocab_uri=None
        ):
    """
    Returns a request dictionary that can be used with the Django test client.
    Per Django documentation, multiple values for a key are provided as a list.
    See: https://docs.djangoproject.com/en/1.8/topics/testing/tools/#making-requests
    """
    form_data_dict = (
        { 'Entity_label':       '%s data ... (%s/_vocab/%s)'%(update, coll_id, vocab_id)
        , 'Entity_comment':     '%s description ... (%s/_vocab/%s)'%(update, coll_id, vocab_id)
        , 'orig_id':            'orig_vocab_id'
        , 'continuation_url':   entitydata_list_type_url(coll_id, "_vocab")
        })
    if vocab_id is not None:
        form_data_dict['entity_id']           = vocab_id
    if vocab_id:
        vocab_url  = recordvocab_url(coll_id=coll_id, vocab_id=vocab_id)
        vocab_help = '%s help for %s in collection %s'%(update, vocab_id, coll_id)
        form_data_dict['entity_id']           = vocab_id
        form_data_dict['Entity_label']        = '%s %s/_vocab/%s'%(update, coll_id, vocab_id)
        form_data_dict['Entity_comment']      = vocab_help
        form_data_dict['Vocab_uri']           = vocab_url or ""
        form_data_dict['orig_id']             = vocab_id
        form_data_dict['orig_type']           = "_vocab"
        form_data_dict['orig_coll']           = coll_id
    if vocab_uri:
        form_data_dict['Vocab_uri']  = vocab_uri
    if orig_id:
        form_data_dict['orig_id']    = orig_id
    if orig_coll:
        form_data_dict['orig_coll']  = orig_coll
    if action:
        form_data_dict['action']     = action
    if cancel:
        form_data_dict['cancel']     = "Cancel"
    elif close:
        form_data_dict['close']      = "Close"
    elif edit:
        form_data_dict['edit']       = "Edit"
    elif copy:
        form_data_dict['copy']       = "Copy"
    elif task:
        form_data_dict[task]         = task
    else:
        form_data_dict['save']       = "Save"
    return form_data_dict

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Tests for EntityData default list view
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                    import settings
from django.db                      import models
from django.http                    import QueryDict
from django.contrib.auth.models     import User
from django.test                    import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions
from django.test.client             import Client

from annalist                       import layout
from annalist                       import message
from annalist.identifiers           import RDF, RDFS, ANNAL

from annalist.models.site           import Site
from annalist.models.collection     import Collection
from annalist.models.recordtype     import RecordType
from annalist.models.recordtypedata import RecordTypeData
from annalist.models.entitydata     import EntityData

from annalist.views.uri_builder             import uri_params, uri_with_params, uri_quote_param
from annalist.views.form_utils.fieldchoice  import FieldChoice

from .AnnalistTestCase import AnnalistTestCase
from .tests import (
    test_layout,
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .init_tests import (
    init_annalist_test_site, init_annalist_test_coll, resetSitedata
    )
from .entity_testutils       import (
    make_message, make_quoted_message,
    site_dir, collection_dir, 
    collection_edit_url,
    continuation_url_param,
    confirm_delete_params,
    collection_create_values,
    site_title,
    create_test_user,
    context_view_field,
    context_list_entities,
    context_list_head_fields,
    context_list_item_fields, context_list_item_field_value
    )
from .entity_testfielddesc import (
    get_field_description, get_bound_field
    )
from .entity_testtypedata import (
    recordtype_dir, 
    recordtype_url,
    recordtype_create_values
    )
from .entity_testentitydata import (
    recorddata_dir,  entitydata_dir,
    entity_url, entitydata_edit_url, entitydata_delete_confirm_url,
    entitydata_list_type_url, entitydata_list_all_url,
    entitydata_value_keys, entitydata_create_values, entitydata_values, 
    entitydata_delete_confirm_form_data,
    entitylist_form_data
    )
from .entity_testsitedata import (
    get_site_types, get_site_types_sorted, get_site_types_linked,
    get_site_lists, get_site_lists_sorted, get_site_lists_linked,
    get_site_views, get_site_views_sorted, get_site_views_linked,
    get_site_list_types, get_site_list_types_sorted,
    get_site_field_groups, get_site_field_groups_sorted, 
    get_site_fields, get_site_fields_sorted, 
    get_site_field_types, get_site_field_types_sorted, 
    )
from .entity_testlistdata import recordlist_url


#   -----------------------------------------------------------------------------
#
#   EntityDefaultListView tests
#
#   -----------------------------------------------------------------------------


class EntityDefaultListViewTest(AnnalistTestCase):
    """
    Tests for default list views
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite  = Site(TestBaseUri, TestBaseDir)
        self.testcoll  = Collection.create(self.testsite, "testcoll", collection_create_values("testcoll"))
        self.testtype  = RecordType.create(self.testcoll, "testtype", recordtype_create_values("testcoll", "testtype"))
        self.testtype2 = RecordType.create(self.testcoll, "testtype2", recordtype_create_values("testcoll", "testtype2"))
        self.testdata  = RecordTypeData.create(self.testcoll, "testtype", {})
        self.testdata2 = RecordTypeData.create(self.testcoll, "testtype2", {})
        e1 = self._create_entity_data("entity1")
        e2 = self._create_entity_data("entity2")
        e3 = self._create_entity_data("entity3")
        e4 = EntityData.create(self.testdata2, "entity4", 
            entitydata_create_values("entity4", type_id="testtype2")
            )
        self.type_ids = get_site_types_linked("testcoll")
        self.type_ids.append(FieldChoice("_type/testtype", 
                label="RecordType testcoll/_type/testtype",
                link=recordtype_url("testcoll", "testtype")
            ))
        self.type_ids.append(FieldChoice("_type/testtype2", 
                label="RecordType testcoll/_type/testtype2",
                link=recordtype_url("testcoll", "testtype2")
            ))
        self.list_ids = get_site_lists_linked("testcoll")
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        # resetSitedata(scope="collections")
        return

    @classmethod
    def setUpClass(cls):
        super(EntityDefaultListViewTest, cls).setUpClass()
        return

    @classmethod
    def tearDownClass(cls):
        super(EntityDefaultListViewTest, cls).tearDownClass()
        #@@checkme@@ resetSitedata()
        return

    #   -----------------------------------------------------------------------------
    #   Helpers
    #   -----------------------------------------------------------------------------

    def _create_entity_data(self, entity_id, update="Entity"):
        "Helper function creates entity data with supplied entity_id"
        e = EntityData.create(self.testdata, entity_id, 
            entitydata_create_values(entity_id, update=update)
            )
        return e    

    #   -----------------------------------------------------------------------------
    #   Form rendering tests
    #   -----------------------------------------------------------------------------

    def test_get_default_all_list(self):
        u = entitydata_list_all_url("testcoll") + "?continuation_url=/xyzzy/"
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, '<h2 class="page-heading">List entities with type information</h2>', html=True)
        self.assertMatch(r.content, r'<input.type="hidden".name="continuation_url".+value="/xyzzy/"/>')
        cont = uri_params({"continuation_url": u})
        tooltip1 = "" # 'title="%s"'%r.context['fields'][0]['field_help']
        tooltip2 = "" # 'title="%s"'%r.context['fields'][1]['field_help']
        tooltip3 = "" # 'title="%s"'%r.context['fields'][2]['field_help']
        rowdata = """
            <div class="tbody row select-row">
              <div class="small-1 columns">
                <input type="checkbox" class="select-box right" name="entity_select"
                       value="testtype/entity1" />
              </div>
              <div class="small-11 columns">
                <div class="row view-listrow">
                  <div class="view-value small-3 columns" %(tooltip1)s>
                    <a href="%(base)s/c/testcoll/d/testtype/entity1/%(cont)s">entity1</a>
                  </div>
                  <div class="view-value small-2 columns" %(tooltip2)s>
                    <a href="/testsite/c/testcoll/d/_type/testtype/%(cont)s">RecordType testcoll/_type/testtype</a>
                  </div>
                  <div class="view-value small-7 columns" %(tooltip3)s>
                    <span>Entity testcoll/testtype/entity1</span>
                  </div>
                </div>
              </div>
            </div>
            """%(
                { 'base':     TestBasePath
                , 'cont':     cont
                , 'tooltip1': tooltip1
                , 'tooltip2': tooltip2
                , 'tooltip3': tooltip3
                }
            )
        # log.info(r.content)
        self.assertContains(r, rowdata, html=True)

        # Test context
        #@@TODO: use new function to create list context for comparison,
        #        like entity view context tests, rather than pick out the fields.
        #
        # self.assertEqual(r.context['title'],            site_title())
        self.assertEqual(r.context['title'],            "List entities with type information - Collection testcoll")
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          None)
        list_choices = r.context['list_choices']
        self.assertEqual(set(list_choices.options),    set(self.list_ids))
        self.assertEqual(list_choices['field_value'],   "Default_list_all")
        self.assertEqual(r.context['continuation_url'], "/xyzzy/")
        # Field descriptions
        head_fields = context_list_head_fields(r.context)
        self.assertEqual(len(head_fields), 1)       # One row of 3 cols..
        self.assertEqual(len(head_fields[0].description['row_field_descs']), 3)
        f0 = context_view_field(r.context, 0, 0)
        f1 = context_view_field(r.context, 0, 1)
        f2 = context_view_field(r.context, 0, 2)
        #  1st field
        self.assertDictionaryMatch(f0, 
            get_bound_field("Entity_id", "", placement="small:0,3")
            )
        #  2nd field
        self.assertDictionaryMatch(f1, 
            get_bound_field("Entity_type", "", 
                placement="small:3,2",
                options=self.type_ids
                )
            )
        # 3rd field
        self.assertDictionaryMatch(f2, 
            get_bound_field("Entity_label", "", placement="small:5,7")
            )
        # Entities and bound fields
        entities = context_list_entities(r.context)
        self.assertEqual(len(entities), 6)
        entity_fields = (
            [ {'entity_type_id': "_type",     'annal:id': "testtype",  'rdfs:label': "RecordType testcoll/_type/testtype"}
            , {'entity_type_id': "_type",     'annal:id': "testtype2", 'rdfs:label': "RecordType testcoll/_type/testtype2"}
            , {'entity_type_id': "testtype",  'annal:id': "entity1",   'rdfs:label': "Entity testcoll/testtype/entity1"}
            , {'entity_type_id': "testtype",  'annal:id': "entity2",   'rdfs:label': "Entity testcoll/testtype/entity2"}
            , {'entity_type_id': "testtype",  'annal:id': "entity3",   'rdfs:label': "Entity testcoll/testtype/entity3"}
            , {'entity_type_id': "testtype2", 'annal:id': "entity4",   'rdfs:label': "Entity testcoll/testtype2/entity4"}
            ])
        field_keys = ('annal:id', 'entity_type_id', 'rdfs:label')
        for eid in range(6):
            item_fields = context_list_item_fields(r.context, entities[eid])
            for fid in range(3):
                item_field = item_fields[fid]
                head_field = head_fields[0].description['row_field_descs'][fid]
                # log.info("Item field: %r"%(item_field,))
                # log.info("Head field: %r"%(head_field,))
                # Check that row field descriptions match corresponding heading feld descriptions
                for fkey in (
                        'field_id', 'field_name', 'field_label', 
                        'field_property_uri', 'field_render_type',
                        'field_placement', 'field_value_type'):
                    self.assertEqual(item_field.description[fkey], head_field[fkey])
                # Check row field values
                fkey = field_keys[fid]
                self.assertEqual(item_field['field_value'],    entity_fields[eid][fkey])
                self.assertEqual(item_field['entity_type_id'], entity_fields[eid]['entity_type_id'])
        return

    def test_get_default_type_list(self):
        u = entitydata_list_type_url("testcoll", "testtype") + "?continuation_url=/xyzzy/"
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<title>List entities - Collection testcoll</title>")
        self.assertContains(r, '<h2 class="page-heading">List entities</h2>', html=True)
        cont = uri_params({"continuation_url": u})
        # cont = ""
        # log.info(r.content)
        tooltip1 = ""
        tooltip2 = ""
        rowdata = """
            <div class="tbody row select-row">
              <div class="small-1 columns">
                <input type="checkbox" class="select-box right" name="entity_select"
                       value="testtype/entity1" />
              </div>
              <div class="small-11 columns">
                <div class="row view-listrow">
                  <div class="view-value small-3 columns" %(tooltip1)s>
                    <a href="%(base)s/c/testcoll/d/testtype/entity1/%(cont)s">entity1</a>
                  </div>
                  <div class="view-value small-9 columns" %(tooltip2)s>
                    <span>Entity testcoll/testtype/entity1</span>
                  </div>
                </div>
              </div>
            </div>
            """%(
                { 'base':     TestBasePath
                , 'cont':     cont
                , 'tooltip1': tooltip1
                , 'tooltip2': tooltip2
                }
            )
        self.assertContains(r, rowdata, html=True)
        # Test context
        # self.assertEqual(r.context['title'],            site_title())
        self.assertEqual(r.context['title'],            "List entities - Collection testcoll")
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "testtype")
        list_choices = r.context['list_choices']
        self.assertEqual(set(list_choices.options),     set(self.list_ids))
        self.assertEqual(list_choices['field_value'],   "Default_list")
        self.assertEqual(r.context['continuation_url'], "/xyzzy/")
        # Fields
        head_fields = context_list_head_fields(r.context)
        self.assertEqual(len(head_fields), 1)       # One row of 2 cols..
        self.assertEqual(len(head_fields[0].description['row_field_descs']), 2)
        f0 = context_view_field(r.context, 0, 0)
        f1 = context_view_field(r.context, 0, 1)
        self.assertDictionaryMatch(f0, 
            get_bound_field("Entity_id", "", placement="small:0,3")
            )
        self.assertDictionaryMatch(f1, 
            get_bound_field("Entity_label", "", placement="small:3,9")
            )
        # Entities
        entities = context_list_entities(r.context)
        self.assertEqual(len(entities), 3)
        field_val = ("entity%d", "Entity testcoll/testtype/entity%d")
        for eid in range(3):
            item_fields = context_list_item_fields(r.context, entities[eid])
            for fid in range(2):
                item_field = item_fields[fid]
                head_field = head_fields[0].description['row_field_descs'][fid]
                for fkey in (
                        'field_id', 'field_name', 'field_label', 
                        'field_property_uri', 'field_render_type',
                        'field_placement', 'field_value_type'):
                    self.assertEqual(item_field.description[fkey], head_field[fkey])
                self.assertEqual(item_field['field_value'], field_val[fid]%(eid+1))
        return

    #   -----------------------------------------------------------------------------
    #   Form response tests
    #   -----------------------------------------------------------------------------

    #   -------- new --------

    def test_post_new_type_entity(self):
        f = entitylist_form_data("new")
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        v = entitydata_edit_url("new", "testcoll", "testtype", view_id="Default_view")
        self.assertIn(v, r['location'])
        return

    def test_post_new_all_entity(self):
        f = entitylist_form_data("new")
        u = entitydata_list_all_url("testcoll")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        v = entitydata_edit_url("new", "testcoll", "Default_type", 
            view_id="Default_view"
            )
        self.assertIn(v, r['location'])
        return

    def test_post_new_type_entity_select_one(self):
        f = entitylist_form_data("new", entities=["testtype/entity1"])
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        v = entitydata_edit_url("new", "testcoll", "testtype", 
            view_id="Default_view"
            )
        self.assertIn(v, r['location'])
        return

    def test_post_new_type_entity_select_many(self):
        f = entitylist_form_data("new", entities=["testtype/entity1", "testtype/entity2"])
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        e1 = u
        e2 = "error_head=%s"%(uri_quote_param(message.INPUT_ERROR),)
        e3 = "error_message="
        self.assertIn(e1, r['location'])
        self.assertIn(e2, r['location'])
        self.assertIn(e3, r['location'])
        return

    #   -------- copy --------

    def test_post_copy_type_entity(self):
        f = entitylist_form_data("copy", entities=["testtype/entity1"])
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        v = entitydata_edit_url("copy", "testcoll", "testtype", "entity1")
        self.assertIn(v, r['location'])
        return

    def test_post_copy_all_entity(self):
        f = entitylist_form_data("copy", entities=["testtype/entity1"])
        u = entitydata_list_all_url("testcoll")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        v = entitydata_edit_url("copy", "testcoll", "testtype", "entity1")
        self.assertIn(v, r['location'])
        return

    def test_post_copy_type_entity_select_none(self):
        f = entitylist_form_data("copy")
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        e1 = u
        e2 = "error_head=%s"%(uri_quote_param(message.INPUT_ERROR),)
        e3 = "error_message=%s"%(uri_quote_param(message.NO_ENTITY_FOR_COPY),)
        self.assertIn(e1, r['location'])
        self.assertIn(e2, r['location'])
        self.assertIn(e3, r['location'])
        return

    def test_post_copy_type_entity_select_many(self):
        f = entitylist_form_data("copy", entities=["testtype/entity1", "testtype/entity2"])
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        e1 = u
        e2 = "error_head=%s"%(uri_quote_param(message.INPUT_ERROR),)
        e3 = "error_message="
        self.assertIn(e1, r['location'])
        self.assertIn(e2, r['location'])
        self.assertIn(e3, r['location'])
        return

    def test_post_copy_type_entity_no_login(self):
        self.client.logout()
        f = entitylist_form_data("copy", entities=["testtype/entity1"])
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   401)
        self.assertEqual(r.reason_phrase, "Unauthorized")
        return

    #   -------- edit --------

    def test_post_edit_type_entity(self):
        f = entitylist_form_data("edit", entities=["testtype/entity1"])
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        v = entitydata_edit_url("edit", "testcoll", "testtype", "entity1")
        self.assertIn(v, r['location'])
        return

    def test_post_edit_all_entity(self):
        f = entitylist_form_data("edit", entities=["testtype/entity1"])
        u = entitydata_list_all_url("testcoll")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        v = entitydata_edit_url("edit", "testcoll", "testtype", "entity1")
        self.assertIn(v, r['location'])
        return

    def test_post_edit_type_entity_select_none(self):
        f = entitylist_form_data("edit")
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        e1 = u
        e2 = "error_head=%s"%(uri_quote_param(message.INPUT_ERROR),)
        e3 = "error_message="
        self.assertIn(e1, r['location'])
        self.assertIn(e2, r['location'])
        self.assertIn(e3, r['location'])
        return

    def test_post_edit_type_entity_select_many(self):
        f = entitylist_form_data("edit", entities=["testtype/entity1", "testtype/entity2"])
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        e1 = u
        e2 = "error_head=%s"%(uri_quote_param(message.INPUT_ERROR),)
        e3 = "error_message="
        self.assertIn(e1, r['location'])
        self.assertIn(e2, r['location'])
        self.assertIn(e3, r['location'])
        return

    def test_post_edit_type_entity_no_login(self):
        self.client.logout()
        f = entitylist_form_data("edit", entities=["testtype/entity1"])
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   401)
        self.assertEqual(r.reason_phrase, "Unauthorized")
        return

    #   -------- delete --------

    def test_post_delete_type_entity(self):
        f = entitylist_form_data("delete", entities=["testtype/entity1"])
        u = entitydata_list_type_url("testcoll", "testtype")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>%s</h3>"%message.CONFIRM_REQUESTED_ACTION)
        msg_text = make_message(message.REMOVE_ENTITY_DATA, 
            type_id="testtype", 
            id="entity1"
            )
        self.assertContains(r, msg_text + ": " + message.ARE_YOU_SURE)
        self.assertContains(r, message.CONFIRM_OR_CANCEL)
        self.assertContains(r,
            '<input type="hidden" name="confirmed_action" value="/testsite/c/testcoll/d/testtype/!delete_confirmed"/>',
            html=True
            )
        self.assertEqual(r.context['action_description'], msg_text)
        self.assertEqual(r.context['confirmed_action'], 
            '/testsite/c/testcoll/d/testtype/!delete_confirmed')
        self.assertEqual(r.context['action_params'], 
            confirm_delete_params(button_id="entity_delete", entity_id="entity1", type_id="testtype")
            )
        self.assertEqual(r.context['cancel_action'], 
            '/testsite/c/testcoll/d/testtype/')
        return

    def test_post_delete_all_entity(self):
        f = entitylist_form_data("delete", entities=["testtype/entity1"])
        u = entitydata_list_all_url("testcoll")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<h3>%s</h3>"%message.CONFIRM_REQUESTED_ACTION)
        # print "**********"
        # print r.content
        # print "**********"
        msg_text = make_message(message.REMOVE_ENTITY_DATA, 
            type_id="testtype", 
            id="entity1"
            )
        self.assertContains(r, msg_text + ": " + message.ARE_YOU_SURE)
        self.assertContains(r, message.CONFIRM_OR_CANCEL)
        self.assertContains(r, 
            '<input type="hidden" name="confirmed_action"  value="/testsite/c/testcoll/d/testtype/!delete_confirmed"/>',
            html=True
            )
        self.assertEqual(r.context['action_description'], msg_text)
        self.assertEqual(r.context['confirmed_action'], 
            '/testsite/c/testcoll/d/testtype/!delete_confirmed')
        self.assertEqual(r.context['action_params'], 
            confirm_delete_params(button_id="entity_delete", entity_id="entity1", type_id=None)
            )
        self.assertEqual(r.context['cancel_action'], 
            '/testsite/c/testcoll/d/')
        return

# End.



"""
am_getargvalue.py - read command value if not already defined
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2013-2014, Graham Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import sys
import getpass
from utils.py3porting import input

# raw_input and getpass:
#   http://packetforger.wordpress.com/2014/03/26/using-pythons-getpass-module/
#   https://docs.python.org/2/library/getpass.html
# Also, for Python 2/3:
#   http://python-future.org/compatible_idioms.html#raw-input

def getarg(args, index):
    """
    Helper to retrieve value from command line args
    """
    return args[index] if index < len(args) else None

def getargvalue(val, prompt):
    """
    Prompt and read value if not defined
    """
    if not val:
        if sys.stdin.isatty():
            val = input(prompt)
        else:
            val = sys.stdin.readline()
            if val[-1] == '\n': val = val[:-1]
    return val

def getsecret(prompt):
    """
    Prompt and read secret value without echo, or read from stdin.
    """
    if sys.stdin.isatty():
        val = getpass.getpass(prompt)
    else:
        val = sys.stdin.readline()
        if val[-1] == '\n': val = val[:-1]
    return val

# End.


"""
Annalist application URL definitions
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

from django.conf.urls                   import url

from annalist.views.home_redirects      import (
    AnnalistHomeView, AnnalistTypeRedirect, AnnalistEntityRedirect
    )
from annalist.views.profile             import ProfileView
from annalist.views.confirm             import ConfirmView
from annalist.views.serverlog           import ServerLogView
from annalist.views.site                import SiteView, SiteActionView
from annalist.views.collection          import CollectionView, CollectionEditView
from annalist.views.recordtypedelete    import RecordTypeDeleteConfirmedView
from annalist.views.recordviewdelete    import RecordViewDeleteConfirmedView
from annalist.views.recordlistdelete    import RecordListDeleteConfirmedView

from annalist.views.entityedit          import GenericEntityEditView
from annalist.views.entitylist          import EntityGenericListView
from annalist.views.entitylistdata      import EntityListDataView
from annalist.views.entitydelete        import EntityDataDeleteConfirmedView

from annalist.views.siteresource        import SiteResourceAccess
from annalist.views.collectionresource  import CollectionResourceAccess
from annalist.views.entityresource      import EntityResourceAccess
from annalist.views.statichack          import serve_pages

from login.login_views                  import LoginUserView, LoginPostView, LogoutUserView
from login.auth_oidc_client             import OIDC_AuthDoneView
from login.auth_django_client           import LocalUserPasswordView

# c - collections
# v - view
# l - list
# d - data/default view
#
# Metadata (using built-in type identifiers, otherwise same pattern as data):
#
# /c/<coll-id>/d/_type/                           list of record types
# /c/<coll-id>/d/_type/<type-id>                  view of type description
# /c/<coll-id>/d/_view/                           list of record views
# /c/<coll-id>/d/_view/<view-id>                  view of view description
# /c/<coll-id>/d/_list/                           list of record lists
# /c/<coll-id>/d/_list/<list-id>                  view of list description
# /c/<coll-id>/d/_field/                          list of field descriptions
# /c/<coll-id>/d/_field/<field-id>                view of field description
#
# Data:
#
# /c/<coll-id>/d/                                 default list of records
# /c/<coll-id>/d/<type-id>/                       default list of records of specified type
# /c/<coll-id>/d/<type-id>/<entity-id>            default view of identified entity
#
# /c/<coll-id>/l/<list-id>/                       specified list of records
# /c/<coll-id>/l/<list-id>/<type-id>              specified list of records of specified type
# /c/<coll-id>/v/<view-id>/<type-id>/<entity-id>  specified view of record
#
# Suffixes /!new, /!copy, /!edit, /!delete, etc. are used for forms that are part of the
# user interface for editing collections and resources, and do not of themselves identify
# persistent resources.

urlpatterns = [

    # Site pages
    url(r'^$',              AnnalistHomeView.as_view(),     name='AnnalistHomeView'),
    url(r'^site/$',         SiteView.as_view(),             name='AnnalistSiteView'),
    url(r'^site/!action$',  SiteActionView.as_view(),       name='AnnalistSiteActionView'),
    url(r'^confirm/$',      ConfirmView.as_view(),          name='AnnalistConfirmView'),
    url(r'^serverlog/$',    ServerLogView.as_view(),        name='AnnalistServerLogView'),

    # Special forms for collection view, customize and type/view/list deletion
    url(r'^c/(?P<coll_id>\w{1,128})/$',
                            CollectionView.as_view(),
                            name='AnnalistCollectionView'),
    url(r'^c/(?P<coll_id>\w{1,128})/!edit$',
                            CollectionEditView.as_view(),
                            name='AnnalistCollectionEditView'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/types/!delete_confirmed$',
                            RecordTypeDeleteConfirmedView.as_view(),
                            name='AnnalistRecordTypeDeleteView'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/views/!delete_confirmed$',
                            RecordViewDeleteConfirmedView.as_view(),
                            name='AnnalistRecordViewDeleteView'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/lists/!delete_confirmed$',
                            RecordListDeleteConfirmedView.as_view(),
                            name='AnnalistRecordListDeleteView'),

    # Default/API access lists and data
    # (these may content negotiate for various formats)
    url(r'^c/(?P<coll_id>\w{1,128})/d/$',
                            EntityGenericListView.as_view(),
                            name='AnnalistEntityDefaultListAll'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<type_id>\w{1,128})/$',
                            EntityGenericListView.as_view(),
                            name='AnnalistEntityDefaultListType'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<type_id>\w{1,128})/!delete_confirmed$',
                            EntityDataDeleteConfirmedView.as_view(),
                            name='AnnalistEntityDataDeleteView'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})/$',
                            GenericEntityEditView.as_view(),
                            name='AnnalistEntityAccessView'),

    # Default edit views
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})/!(?P<action>copy)$',
                            GenericEntityEditView.as_view(),
                            name='AnnalistEntityDefaultDataView'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})/!(?P<action>edit)$',
                            GenericEntityEditView.as_view(),
                            name='AnnalistEntityDefaultDataView'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})/!(?P<action>view)$',
                            GenericEntityEditView.as_view(),
                            name='AnnalistEntityDefaultDataView'),

    # JSON list views without list_id specified
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<list_ref>entity_list.[\w]{1,32})$',
                            EntityListDataView.as_view(),
                            name='AnnalistEntityListDataAll'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<type_id>\w{1,128})/(?P<list_ref>entity_list.[\w]{1,32})$',
                            EntityListDataView.as_view(),
                            name='AnnalistEntityListDataType'),

    # Redirect type/entity URIs without trailing '/'
    # (Note these cannot match JSON resource names as '.' is not matched here)
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<type_id>\w{1,128})$',
                            AnnalistTypeRedirect.as_view(),
                            name='AnnalistTypeRedirect'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})$',
                            AnnalistEntityRedirect.as_view(),
                            name='AnnalistEntityRedirect'),

    # Specified list views
    url(r'^c/(?P<coll_id>\w{1,128})/l/$',
                            EntityGenericListView.as_view(),
                            name='AnnalistEntityDefaultList'),
    url(r'^c/(?P<coll_id>\w{1,128})/l/(?P<list_id>\w{1,128})/$',
                            EntityGenericListView.as_view(),
                            name='AnnalistEntityGenericList'),
    url(r'^c/(?P<coll_id>\w{1,128})/l/(?P<list_id>\w{1,128})/(?P<type_id>\w{1,128})/$',
                            EntityGenericListView.as_view(),
                            name='AnnalistEntityGenericList'),

    # JSON specified list views
    url(r'^c/(?P<coll_id>\w{1,128})/l/(?P<list_ref>entity_list.[\w]{1,32})$',
                            EntityListDataView.as_view(),
                            name='AnnalistEntityListDataAll'),
    url(r'^c/(?P<coll_id>\w{1,128})/l/(?P<list_id>\w{1,128})/(?P<list_ref>entity_list.[\w]{1,32})$',
                            EntityListDataView.as_view(),
                            name='AnnalistEntityListDataAll'),
    url(r'^c/(?P<coll_id>\w{1,128})/l/(?P<list_id>\w{1,128})/(?P<type_id>\w{1,128})/(?P<list_ref>entity_list.[\w]{1,32})$',
                            EntityListDataView.as_view(),
                            name='AnnalistEntityListDataType'),

    # Specified entity edit/view forms
    url(r'^c/(?P<coll_id>\w{1,128})/v/(?P<view_id>\w{1,128})/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})/$',
                            GenericEntityEditView.as_view(),
                            name='AnnalistEntityDataView'),
    url(r'^c/(?P<coll_id>\w{1,128})/v/(?P<view_id>\w{1,128})/(?P<type_id>\w{1,128})/!(?P<action>new)$',
                            GenericEntityEditView.as_view(),
                            name='AnnalistEntityNewView'),
    url(r'^c/(?P<coll_id>\w{1,128})/v/(?P<view_id>\w{1,128})/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})/!(?P<action>copy)$',
                            GenericEntityEditView.as_view(),
                            name='AnnalistEntityEditView'),
    url(r'^c/(?P<coll_id>\w{1,128})/v/(?P<view_id>\w{1,128})/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})/!(?P<action>edit)$',
                            GenericEntityEditView.as_view(),
                            name='AnnalistEntityEditView'),
    url(r'^c/(?P<coll_id>\w{1,128})/v/(?P<view_id>\w{1,128})/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})/!(?P<action>view)$',
                            GenericEntityEditView.as_view(),
                            name='AnnalistEntityEditView'),

    # Named resource access (metadata, context, attachments, etc.)
    url(r'^site/(?P<resource_ref>[\w.-]{1,250})$',
                            SiteResourceAccess.as_view(),
                            name='AnnalistSiteResourceAccess'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<resource_ref>[\w.-]{1,250})$',
                            CollectionResourceAccess.as_view(),
                            name='AnnalistCollectionResourceAccess'),
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})/(?P<resource_ref>[\w.-]{1,250})$',
                            EntityResourceAccess.as_view(),
                            name='AnnalistEntityResourceAccess'),

    # Entity resource data access with specified view
    url(r'^c/(?P<coll_id>\w{1,128})/v/(?P<view_id>\w{1,128})/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})/(?P<resource_ref>[\w.-]{1,250})$',
                            EntityResourceAccess.as_view(),
                            name='AnnalistEntityViewAccess'),

    # Access supporting application pages in same collection
    url(r'^c/(?P<coll_id>\w{1,128})/p/(?P<page_ref>[\w/.-]{1,250})$',
                            serve_pages),
    ] # End of urlpatterns

# Login-related view URLs

urlpatterns += [
    url(r'^login/$',        LoginUserView.as_view(),            name='LoginUserView'),
    url(r'^login_post/$',   LoginPostView.as_view(),            name='LoginPostView'),
    url(r'^login_local/$',  LocalUserPasswordView.as_view(),    name='LocalUserPasswordView'),
    url(r'^login_done/',    OIDC_AuthDoneView.as_view(),        name='OIDC_AuthDoneView'),
    url(r'^profile/$',      ProfileView.as_view(),              name='AnnalistProfileView'),
    url(r'^logout/$',       LogoutUserView.as_view(),           name='LogoutUserView'),
    # Info view...
    # url(r'^c/(?P<coll_id>_annalist_site)/d/(?P<type_id>_info)/(?P<entity_id>about)/$',
    url(r'^c/(?P<coll_id>\w{1,128})/d/(?P<type_id>\w{1,128})/(?P<entity_id>\w{1,128})/$',
                            GenericEntityEditView.as_view(),    name='AnnalistInfoView'),
    ]

# End.



"""
Definitions to support porting to Python 3

The intent is to abstract here those API calls used by Annalist that are 
sensitive to string vs unicode parameters, and changes to the Python library
structure between Python versions 2 and 3.  It also serves to identify code
features for which there isn't a natural idiom that works across versions.

To use this module, include something like this in the original source:

    from utils.py3porting import is_string, to_unicode, urljoin
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

try:
    # Python3
    from urllib.parse       import (
        urlparse, urljoin, 
        urlsplit, urlunsplit, 
        quote, unquote
        )
    from urllib.request     import urlopen, Request, pathname2url
    from urllib.error       import HTTPError
except ImportError:
    # Python2
    from urlparse           import urlparse, urljoin, urlsplit, urlunsplit
    from urllib2            import urlopen, Request, HTTPError
    from urllib             import quote, unquote, pathname2url

import six
from six import text_type, StringIO, BytesIO, iteritems, iterlists
from six.moves import input

def is_string(val):
    """
    Is the supplied value a string or unicode string?

    See: https://stackoverflow.com/a/33699705/324122
    """
    return isinstance(val, (str, u"".__class__))

def to_unicode(val):
    """
    Converts a supplied string value to Unicode text
    """
    return text_type(val)

def text_to_str(ustr):
    """
    Return string value for supplied Unicode
    """
    return str(ustr)

def bytes_to_str(bstr):
    """
    Return string value for supplied bytes
    """
    # return bstr.encode('ascii', 'ignore')
    if six.PY3:
        return bstr.decode('ascii', 'ignore')
    return bstr

def text_to_bytes(ustr):
    """
    Return bytes value for supplied string.
    The intent is that the string may be an ASCII or unicode string, but not
    something that has already been encoded.
    """
    return ustr.encode('utf-8', 'ignore')

def bytes_to_unicode(bstr):
    """
    Return Unicode value for supplied (UTF-8 encoding) bytes.
    """
    return bstr.decode('utf-8')

str_space = text_to_str(' ')

def write_bytes(file, text):
    """
    Write supplied string to file as bytes
    """
    file.write(text_to_bytes(text))
    return

def isoformat_space(datetime):
    """
    Return ISO-formatted date with space to separate date and time.
    """
    return datetime.isoformat(str_space)

def get_message_type(msg_info):
    """
    Return content type of result returned by urlopen.

    The message info value returned by Python2's urllib2.urlopen has long been
    deprecated.  The newer methods return info() as an `email.message.Message`
    value, whose corresponding content-type method is `get_content_type`.
    """
    try:
        # Python3
        return msg_info.get_content_type()
    except AttributeError:
        # Python2
        return msg_info.gettype()

# End.



"""
Tests for repeat group rendering
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import sys
import os
import unittest
import re

from django.test.client                 import Client

from annalist.models.site               import Site
from annalist.models.collection         import Collection
from annalist.models.recordtype         import RecordType
from annalist.models.recordtypedata     import RecordTypeData
from annalist.models.recordview         import RecordView
from annalist.models.recordfield        import RecordField
from annalist.models.entitydata         import EntityData

from .AnnalistTestCase import AnnalistTestCase
from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .init_tests import (
    init_annalist_test_site,
    init_annalist_test_coll,
    resetSitedata
    )
from .entity_testfielddesc import (
    get_field_description, get_bound_field
    )
from .entity_testutils import (
    collection_create_values,
    create_test_user,
    context_view_field,
    context_bind_fields,
    context_field_row
    )
from .entity_testtypedata import (
    recordtype_create_values, 
    )
from .entity_testentitydata import (
    entity_url, entitydata_edit_url, 
    specified_view_context_data,
    default_fields
    )

class RepeatGroupRenderingTest(AnnalistTestCase):

    def setUp(self):
        # Set up basic site data
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection.create(self.testsite, "testcoll", collection_create_values("testcoll"))
        self.testtype = RecordType.create(self.testcoll, "testtype", recordtype_create_values("testtype"))
        self.testdata = RecordTypeData.create(self.testcoll, "testtype", {})
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        # resetSitedata(scope="collections")
        return


    @classmethod
    def setUpClass(cls):
        super(RepeatGroupRenderingTest, cls).setUpClass()
        return

    @classmethod
    def tearDownClass(cls):
        super(RepeatGroupRenderingTest, cls).tearDownClass()
        resetSitedata(scope="collections")
        return

    # Support methods

    def _create_testview(self):
        testview = RecordView.create(self.testcoll, "testview",
            { 'annal:type':             "annal:View"
            , 'annal:uri':              "test:testtype"
            , 'rdfs:label':             "Test view label"
            , 'rdfs:comment':           "Test view comment"
            , 'annal:view_entity_type': "testtype"
            , 'annal:view_fields':
              [ { 'annal:field_id':         "Entity_id"
                , 'annal:field_placement':  "small:0,12;medium:0,6"
                }
              , { 'annal:field_id':         "Test_repeat_field"
                , 'annal:field_placement':  "small:0,12"
                }
              ]
            })
        self.assertTrue(testview is not None)
        return testview

    def _create_testrepeatfield(self, label_add=None, label_delete=None):
        testrepeatfield = RecordField.create(self.testcoll, "Test_repeat_field",
            { "annal:type":                 "annal:Field"
            , "rdfs:label":                 "Test repeat field label"
            , "rdfs:comment":               "Test repeat field comment"
            , "annal:field_render_type":    "Group_Seq_Row"
            , "annal:field_value_mode":     "Value_direct"
            , "annal:field_value_type":     "annal:Field_group"
            , "annal:field_entity_type":    "test:testtype"
            , "annal:placeholder":          "(test repeat field)"
            , "annal:property_uri":         "test:repeat_fields"
            , "annal:field_placement":      "small:0,12"
            , "annal:field_fields": 
                [ { "annal:field_id":   "Entity_comment" }
                ]
            , "annal:repeat_label_add":     label_add
            , "annal:repeat_label_delete":  label_delete
            })
        self.assertTrue(testrepeatfield is not None)
        return testrepeatfield

    def _create_testentity(self):
        testentity = EntityData.create(self.testdata, "testentity",
            { "annal:type":         "test:testtype"
            , "test:repeat_fields": []
            })
        self.assertTrue(testentity is not None)
        return testentity

    def _create_repeat_group_view_context(self, 
            entity_id=None,
            action=None,
            label_add=None,
            label_delete=None,
            field_value=None,
            continuation_url=None
        ):
        view_fields = (
            [ context_field_row(
                get_bound_field("Entity_id",         entity_id,
                    placement="small:0,12;medium:0,6"
                    ),
                )
            , get_bound_field("Test_repeat_field",   field_value,
                placement="small:0,12",
                group_add_label=label_add,
                group_delete_label=label_delete,
                )
            ])
        context_dict = specified_view_context_data(
            coll_id="testcoll", type_id="testtype", 
            view_id="testview", view_heading="Test view label",
            entity_id=entity_id,
            entity_label="",
            view_fields=view_fields,
            action=action, 
            continuation_url=continuation_url
            )
        return context_dict

    # Tests

    def test_RenderRepeatGroup(self):
        # Create view with repeat-field and repeat-group
        testview        = self._create_testview()
        testrepeatfield = self._create_testrepeatfield(
            label_add="Add group",
            label_delete="Remove group")

        # Render view
        u = entitydata_edit_url("new", "testcoll", "testtype", view_id="testview")
        response = self.client.get(u)
        self.assertEqual(response.status_code,   200)
        self.assertEqual(response.reason_phrase, "OK")

        # Check render context
        expect_context = self._create_repeat_group_view_context(
            entity_id="00000001",
            action="new",
            label_add="Add group",
            label_delete="Remove group",
            field_value="",
            continuation_url=""
            )
        actual_context = context_bind_fields(response.context)
        self.assertEqual(len(response.context['fields']), 2)
        self.assertDictionaryMatch(actual_context, expect_context)

        # Test rendered result
        f0 = context_view_field(response.context, 0, 0)
        f1 = context_view_field(response.context, 1, 0)
        field_vals = default_fields(
            coll_id="testcoll", type_id="testtype", entity_id="00000001",
            tooltip1=f0['field_tooltip'],
            )
        formrow1 = """
            <div class="small-12 medium-6 columns" title="%(tooltip1)s">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Id</span>
                </div>
                <div class="%(input_classes)s">
                    <input type="text" size="64" name="entity_id" 
                           placeholder="(entity id)" value="00000001"/>
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        formrow2a = """
            <div class="%(group_label_classes)s">
              <span>Test repeat field label</span>
            </div>
            """%field_vals(width=12)
        formrow2b = """
            <div class="%(group_row_head_classes)s">
              <div class="row">
                <div class="small-1 columns">
                  &nbsp;
                </div>
                <div class="small-11 columns">
                  <div class="edit-grouprow col-head row">
                    <div class="%(col_head_classes)s">
                      <span>Comment</span>
                    </div>
                  </div>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow2c = """
            <div class="%(group_buttons_classes)s">
              <div class="row">
                <div class="small-1 columns">
                  &nbsp;
                </div>
                <div class="small-11 columns">
                  <input type="submit" name="Test_repeat_field__remove" value="Remove group" />
                  <input type="submit" name="Test_repeat_field__add"    value="Add group" />
                  <input type="submit" name="Test_repeat_field__up"     value="Move &#x2b06;" />
                  <input type="submit" name="Test_repeat_field__down"   value="Move &#x2b07;" />
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        # log.info(r.content)
        self.assertContains(response, formrow1,  html=True)
        self.assertContains(response, formrow2a, html=True)
        self.assertContains(response, formrow2b, html=True)
        self.assertContains(response, formrow2c, html=True)
        return

    def test_RenderRepeatGroup_no_labels(self):
        # Create view with repeat-field and repeat-group
        testview        = self._create_testview()
        Test_repeat_field = self._create_testrepeatfield(
            label_add=None,
            label_delete="")

        # Render view
        u = entitydata_edit_url("new", "testcoll", "testtype", view_id="testview")
        response = self.client.get(u)
        self.assertEqual(response.status_code,   200)
        self.assertEqual(response.reason_phrase, "OK")

        # Check render context
        expect_context = self._create_repeat_group_view_context(
            entity_id="00000001",
            action="new",
            label_add="Add Test_repeat_field",
            label_delete="Remove Test_repeat_field",
            field_value="",
            continuation_url=""
            )
        actual_context = context_bind_fields(response.context)
        self.assertEqual(len(response.context['fields']), 2)
        self.assertDictionaryMatch(actual_context, expect_context)
        return

    def test_RenderRepeatGroup_view_no_values(self):
        # Create view with repeat-field and repeat-group
        testview        = self._create_testview()
        testrepeatfield = self._create_testrepeatfield(
            label_add="Add group",
            label_delete="Remove group")
        testentity      = self._create_testentity()
        # Render view
        u = entitydata_edit_url(
            "view", "testcoll", "testtype", entity_id="testentity", view_id="testview"
            )
        v = entity_url("testcoll", "testtype", entity_id="testentity")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test rendered result
        cont_uri_param = "?continuation_url="+u
        field_vals = default_fields(
            coll_id="testcoll", type_id="testtype", entity_id="00000001",
            view_url=v, cont_uri_param=cont_uri_param,
            tooltip1="",
            tooltip2="",
            )
        formrow1 = """
            <div class="small-12 medium-6 columns" %(tooltip1)s>
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Id</span>
                </div>
                <div class="%(input_classes)s">
                  <a href="%(view_url)s%(cont_uri_param)s">testentity</a>
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        formrow2 = """
            <div class="grouprow row">
              <div class="%(group_label_classes)s">
                <span>Test repeat field label</span>
              </div>
              <div class="group-placeholder %(group_row_body_classes)s">
                <span>(None specified)</span>
              </div>
            </div>
            """%field_vals(width=12)
        # log.info(r.content) #@@
        self.assertContains(r, formrow1,  html=True)
        self.assertContains(r, formrow2, html=True)
        return

# End.

if __name__ == "__main__":
    # import django
    # django.setup()  # Needed for template loader
    # Runtests in this module
    # runner = unittest.TextTestRunner(verbosity=2)
    # tests = unittest.TestSuite()
    # tests  = getSuite(select=sel)
    # if tests: runner.run(tests)
    unittest.main()



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Tests for user profile page (also tests user identity access)
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                import settings
from django.db                  import models
from django.http                import QueryDict
from django.core.urlresolvers   import resolve, reverse
from django.contrib.auth.models import User
from django.test                import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions
from django.test.client         import Client

from annalist                   import layout
from annalist.identifiers       import RDF, RDFS, ANNAL
from annalist.models.site       import Site
from annalist.views.profile     import ProfileView

from .AnnalistTestCase import AnnalistTestCase
from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .init_tests import (
    init_annalist_test_site, 
    init_annalist_test_coll,
    resetSitedata
    )
from .entity_testutils import site_title

class UserProfileTest(AnnalistTestCase):
    """
    Tests for user profile page
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        return

    def tearDown(self):
        return

    @classmethod
    def setUpClass(cls):
        super(UserProfileTest, cls).setUpClass()
        return

    @classmethod
    def tearDownClass(cls):
        super(UserProfileTest, cls).tearDownClass()
        # @@checkme@@ resetSitedata(scope="all")
        return

    def test_UserProfileTest(self):
        self.assertEqual(ProfileView.__name__, "ProfileView", "Check ProfileView class name")
        return


#   -----------------------------------------------------------------------------
#
#   UserProfileView tests
#
#   -----------------------------------------------------------------------------

class UserProfileViewTest(AnnalistTestCase):
    """
    Tests for Site views
    """

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.user = User.objects.create_user(
            'testuser', 'user@test.example.com', 'testpassword',
            first_name="Test", last_name="User"
            )
        self.user.save()
        self.client     = Client(HTTP_HOST=TestHost)
        self.uri        = reverse("AnnalistProfileView")
        self.homeuri    = reverse("AnnalistHomeView")
        return

    def tearDown(self):
        return

    def test_UserProfileViewTest(self):
        self.assertEqual(ProfileView.__name__, "ProfileView", "Check ProfileView class name")
        return

    def test_get(self):
        # @@TODO: use reference to self.client, per 
        # https://docs.djangoproject.com/en/dev/topics/testing/tools/#default-test-client
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        u = reverse("AnnalistProfileView")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        row1 = """
            <div class="row view-value-row">
              <div class="view-label small-12 medium-2 columns">
                <span>User</span>
              </div>
              <div class="view-value small-12 medium-10 columns">
                <span>testuser</span>
              </div>
            </div>
            """
        row2 = """
            <div class="row view-value-row">
              <div class="view-label small-12 medium-2 columns">
                <span>Name</span>
              </div>
              <div class="view-value small-12 medium-10 columns">
                <span>Test User</span>
              </div>
            </div>
            """
        row3 = """
            <div class="row view-value-row">
              <div class="view-label small-12 medium-2 columns">
                <span>Email</span>
              </div>
              <div class="view-value small-12 medium-10 columns">
                <span><a href="mailto:user@test.example.com">user@test.example.com</a></span>
              </div>
            </div>
            """

        # log.info(r.content)
        self.assertContains(r, site_title("<title>%s</title>"))
        self.assertContains(r, row1, html=True)
        self.assertContains(r, row2, html=True)
        self.assertContains(r, row3, html=True)
        return

    def test_get_no_login(self):
        u = reverse("AnnalistProfileView")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "Found")
        self.assertEqual(r.content,       b"")
        self.assertIn("/testsite/login/", r['location'])
        self.assertIn("continuation_url=/testsite/", r['location'])
        return

# End.



"""
Record field data functions to support entity data testing
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os

from django.conf                import settings
from django.http                import QueryDict
from django.utils.http          import urlquote, urlunquote
from django.core.urlresolvers   import resolve, reverse

from utils.py3porting           import urljoin

from annalist.util              import valid_id, extract_entity_id
from annalist.identifiers       import RDF, RDFS, ANNAL
from annalist                   import layout

from annalist.views.fields.render_placement import (
    get_placement_classes
    )

from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from .entity_testutils import (
    collection_dir, 
    collection_edit_url,
    collection_entity_view_url,
    site_title,
    context_field_row
    )
from .entity_testentitydata     import entitydata_list_type_url
from .entity_testfielddesc      import get_field_description, get_bound_field

#   -----------------------------------------------------------------------------
#
#   Directory generating functions
#
#   -----------------------------------------------------------------------------

def recordtype_dir(coll_id="testcoll", type_id="testtype"):
    return collection_dir(coll_id) + layout.COLL_TYPE_PATH%{'id': type_id} + "/"

#   -----------------------------------------------------------------------------
#
#   URI generating functions
#
#   -----------------------------------------------------------------------------

#   These use the Django `reverse` function so they correspond to
#   the declared URI patterns.

def recordtype_coll_url(site, coll_id="testcoll", type_id="testtype"):
    return urljoin(
        site._entityurl,
        layout.SITE_COLL_PATH%{'id': coll_id} + "/" + 
        layout.COLL_TYPE_PATH%{'id': type_id} + "/"
        )

def recordtype_url(coll_id="testcoll", type_id="testtype"):
    """
    URL for record type description data; also view using default entity view
    """
    if not valid_id(type_id):
        return None
    return collection_entity_view_url(coll_id=coll_id, type_id="_type", entity_id=type_id)

def recordtype_edit_url(action=None, coll_id=None, type_id=None):
    """
    URI for record type description editing view
    """
    viewname = ( 
        'AnnalistEntityDataView'        if action == "view"   else
        'AnnalistEntityNewView'         if action == "new"    else
        'AnnalistEntityEditView'        if action == "copy"   else
        'AnnalistEntityEditView'        if action == "edit"   else
        'AnnalistRecordTypeDeleteView'  if action == "delete" else
        'unknown'
        )
    kwargs = {'coll_id': coll_id}
    if action != "delete":
        kwargs.update({'action': action, 'type_id': "_type", 'view_id': "Type_view"})
    if type_id:
        if valid_id(type_id):
            kwargs.update({'entity_id': type_id})
        else:
            kwargs.update({'entity_id': "___"})
    url = reverse(viewname, kwargs=kwargs)
    return url

#   -----------------------------------------------------------------------------
#
#   ----- RecordType data
#
#   -----------------------------------------------------------------------------

def recordtype_value_keys(type_uri=False, supertype_uris=False):
    ks = set(
        [ 'annal:id', 'annal:type_id'
        , 'annal:type'
        , 'annal:url'
        , 'rdfs:label', 'rdfs:comment'
        , 'annal:type_view', 'annal:type_list'
        ])
    if type_uri:
        ks.add('annal:uri')
    if supertype_uris:
        ks.add('annal:supertype_uri')
    return ks

def recordtype_load_keys(type_uri=False, supertype_uris=False):
    return (
        recordtype_value_keys(type_uri=type_uri, supertype_uris=supertype_uris) | 
        {'@id', '@type', '@context'}
        )

def recordtype_create_values(
        coll_id="testcoll", type_id="testtype", update="RecordType",
        type_uri=None, supertype_uris=None
        ):
    """
    Entity values used when creating a record type entity
    """
    d = (
        { 'annal:type':         "annal:Type"
        , 'rdfs:label':         "%s %s/%s/%s"%(update, coll_id, "_type", type_id)
        , 'rdfs:comment':       '%s coll %s, type %s, entity %s'%(update, coll_id, "_type", type_id)
        , 'annal:type_view':    "_view/Default_view"
        , 'annal:type_list':    "_list/Default_list"
        })
    if type_uri:
        d['annal:uri'] = type_uri
        if supertype_uris is not None:
            d['annal:supertype_uri'] = (
                [ { '@id': st } for st in supertype_uris ]
                )
        else:
            d['annal:supertype_uri'] = (
                [ { '@id': type_uri+"/super1" }
                , { '@id': type_uri+"/super2" }
                ])
    return d

def recordtype_values(
        coll_id="testcoll", type_id="testtype", type_uri=None,
        update="RecordType", hosturi=TestHostUri):
    type_url = recordtype_url(coll_id=coll_id, type_id=type_id)
    d = recordtype_create_values(
        coll_id, type_id, update=update, type_uri=type_uri
        ).copy()
    d.update(
        { 'annal:id':       type_id
        , 'annal:type_id':  "_type"
        , 'annal:url':      type_url
        })
    return d

def recordtype_read_values(
        coll_id="testcoll", type_id="testtype", 
        update="RecordType", hosturi=TestHostUri):
    d = recordtype_values(coll_id, type_id,
        update=update,
        hosturi=hosturi
        ).copy()
    d.update(
        { '@id':            layout.COLL_BASE_TYPE_REF%{'id': type_id}
        , '@type':          ["annal:Type"]
        , '@context':       [{"@base": "../../"}, "../../coll_context.jsonld"]
        })
    return d

#   -----------------------------------------------------------------------------
#
#   ----- Entity data in recordtype view
#
#   -----------------------------------------------------------------------------

def type_view_context_data(action=None,
        coll_id="testcoll",
        type_type_id=layout.TYPE_TYPEID, orig_type=None,
        type_entity_id="", orig_id=None, type_ids=[],
        type_label=None,
        type_descr=None,
        type_uri=None,
        type_supertype_uris=[],
        type_view="_view/Default_view",
        type_list="_list/Default_list",
        type_aliases=[],
        record_type="annal:Type",
        update="RecordType",
        continuation_url=None
    ):
    if type_uri is None:
        type_uri = ""
        if type_entity_id:
            type_uri = recordtype_url(coll_id=coll_id, type_id=type_entity_id)
    if type_label is None:
        if type_entity_id:
            type_label = "%s %s/%s/%s"%(update, coll_id, type_type_id, type_entity_id)
        elif orig_id:
            type_label = "%s %s/%s/%s"%(update, coll_id, type_type_id, orig_id)
        else:
            type_label = "%s data ... (%s/%s)"%(update, coll_id, type_entity_id)
    if type_uri is None:
       type_uri = recordtype_url(coll_id=coll_id, type_id=type_entity_id)
    if continuation_url is None:
        continuation_url = entitydata_list_type_url(coll_id, type_type_id)
    view_heading = "Type definition"
    view_title   = (
        "%s - %s - Collection %s"%(type_label, view_heading, coll_id) if type_label
        else
        "%s - Collection %s"%(view_heading, coll_id)
        )
    context_dict = (
        { 'title':              view_title
        , 'heading':            view_heading
        , 'coll_id':            coll_id
        , 'type_id':            type_type_id
        , 'view_id':            "Type_view"
        , 'entity_id':          type_entity_id or ""
        , 'orig_type':          orig_type or type_type_id
        , 'record_type':        record_type
        , 'continuation_url':   continuation_url
        , 'fields':
          [ context_field_row(
              get_bound_field("Type_id",           type_entity_id),         # 0 (0,0)
              )
          , context_field_row(
              get_bound_field("Type_label",        type_label)              # 1 (1,0)
              )
          , context_field_row(
              get_bound_field("Type_comment",      type_descr)              # 2 (2,0)
              )
          , context_field_row(
              get_bound_field("Type_uri",          type_uri),               # 3 (3,0)
              )
          , get_bound_field("Type_supertype_uris", type_supertype_uris)     # 4 (4)  
          , context_field_row(
              get_bound_field("Type_view",         type_view),              # 5 (5,0)
              get_bound_field("Type_list",         type_list)               # 6 (5,1)
              )
          , get_bound_field("Type_aliases",        type_aliases)            # 7 (6)
          ]
        })
    if orig_id is not None:
        context_dict['orig_id']     = orig_id
    elif action != "new":
        context_dict['orig_id']     = type_entity_id
    if action:  
        context_dict['action']      = action
    return context_dict

def type_view_form_data(action=None, 
        coll_id="testcoll", orig_coll=None,
        type_type_id="_type", orig_type=None,
        type_entity_id="", orig_id=None, type_entity_uri=None,
        cancel=None, close=None, edit=None, copy=None, task=None,
        add_view_field=None, open_view=None, customize=None,
        update="RecordType",
        ):
    """
    Returns a request dictionary that can be used with the Django test client.
    Per Django documentation, multiple values for a key are provided as a list.
    See: https://docs.djangoproject.com/en/1.8/topics/testing/tools/#making-requests

    Note: historically, some tests use Type_view to display non-type data, 
    hence explicit type_type_id and type_entity_id parameters.
    """
    form_data_dict = (
        { 'Type_label':         '%s data ... (%s/%s)'%(update, coll_id, type_entity_id)
        , 'Type_comment':       '%s description ... (%s/%s)'%(update, coll_id, type_entity_id)
        , 'continuation_url':   entitydata_list_type_url(coll_id, type_type_id)
        })
    if action:
        form_data_dict['action']        = action
    if type_type_id:
        form_data_dict['entity_type']   = "_type/"+type_type_id
        form_data_dict['orig_type']     = extract_entity_id(type_type_id)
    if type_entity_id is not None:
        form_data_dict['entity_id']     = type_entity_id
        form_data_dict['orig_id']       = type_entity_id
    if type_entity_id and type_type_id:
        entity_url  = recordtype_url(coll_id=coll_id, type_id=type_entity_id)
        form_data_dict['entity_id']     = type_entity_id
        form_data_dict['Type_uri']      = entity_url or ""
        form_data_dict['Type_view']     = "_view/Default_view"
        form_data_dict['Type_list']     = "_list/Default_list"
        form_data_dict['orig_coll']     = coll_id
    if orig_coll:
        form_data_dict['orig_coll']     = orig_coll
    if orig_type:
        form_data_dict['orig_type']     = orig_type
    if orig_id:
        form_data_dict['orig_id']       = orig_id
    label_id = type_entity_id or orig_id
    if label_id and type_type_id:
        form_data_dict['Type_label']    = (
            '%s %s/%s/%s'%(update, coll_id, type_type_id, label_id)
            )
        form_data_dict['Type_comment']  = (
            '%s coll %s, type %s, entity %s'%(update, coll_id, type_type_id, label_id)
            )
    if type_entity_uri:
        form_data_dict['Type_uri']                                   = type_entity_uri
        type_uri_rstrip = type_entity_uri.rstrip()
        form_data_dict['Type_supertype_uris__0__Type_supertype_uri'] = type_uri_rstrip+"/super1"
        form_data_dict['Type_supertype_uris__1__Type_supertype_uri'] = type_uri_rstrip+"/super2"
    if cancel:
        form_data_dict['cancel']            = "Cancel"
    elif close:
        form_data_dict['close']             = "Close"
    elif edit:
        form_data_dict['edit']              = "Edit"
    elif copy:
        form_data_dict['copy']              = "Copy"
    elif add_view_field:
        form_data_dict['add_view_field']    = add_view_field
    elif open_view:
        form_data_dict['open_view']         = open_view
    elif customize:
        form_data_dict['customize']         = customize
    elif task:
        form_data_dict[task]                = task
    else:
        form_data_dict['save']      = "Save"
    return form_data_dict

#   -----------------------------------------------------------------------------
#
#   ----- Recordtype delete confirmation form data
#
#   -----------------------------------------------------------------------------

def recordtype_delete_form_data(type_id=None, list_id="Default_list"):
    return (
        { 'list_choice':        "_list/"+list_id
        , 'continuation_url':   ""
        , 'search_for':         ""
        , 'entity_select':      ["_type/%s"%(type_id)]
        , 'delete':             "Delete"
        })

def recordtype_delete_confirm_form_data(type_id=None):
    return (
        { 'typelist':    type_id,
          'type_delete': 'Delete'
        })

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
This module implements a class that is used for rendering a bound field, given
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import traceback

from annalist.views.fields.find_renderers   import (
    get_label_renderer,
    get_view_renderer,
    get_edit_renderer, 
    get_label_view_renderer,
    get_label_edit_renderer, 
    get_col_head_renderer, 
    get_col_head_view_renderer, 
    get_col_head_edit_renderer, 
    get_col_view_renderer,
    get_col_edit_renderer,
    get_value_mapper
    )

class FieldRenderer(object):
    """
    This class represents a value renderer that is bound to a specific field value.

    Methods are referenced from Django templates by a {% include ... %} directive.
    """

    def __init__(self, field_render_type, field_value_mode):
        """
        Initialize a bound field renderer object.
        """
        self._field_render_type = field_render_type
        self._field_value_mode  = field_value_mode
        self._value_mapper      = get_value_mapper(field_render_type)
        self._value_renderer    = (
            { 'label':         get_label_renderer(        field_render_type, field_value_mode)
            , 'view':          get_view_renderer(         field_render_type, field_value_mode)
            , 'edit':          get_edit_renderer(         field_render_type, field_value_mode)
            , 'label_view':    get_label_view_renderer(   field_render_type, field_value_mode)
            , 'label_edit':    get_label_edit_renderer(   field_render_type, field_value_mode)
            , 'col_head':      get_col_head_renderer(     field_render_type, field_value_mode)
            , 'col_head_view': get_col_head_view_renderer(field_render_type, field_value_mode)
            , 'col_head_edit': get_col_head_edit_renderer(field_render_type, field_value_mode)
            , 'col_view':      get_col_view_renderer(     field_render_type, field_value_mode)
            , 'col_edit':      get_col_edit_renderer(     field_render_type, field_value_mode)
            })
        return

    def __str__(self):
        return (
            "FieldRenderer(render_type=%s, value_mode=%s)"%
            (self._field_render_type, self._field_value_mode)
            )

    def value_mapper(self):
        return self._value_mapper

    def renderer(self, mode):
        """
        Return a renderer for a specified mode.

        The renderer returned is accessed from a template using a {% include %} directive,
        which can accept an object with a render method.  A context value is supplied by 
        the Django template processor at the point of invocation.  This mechanism is used
        in the "render_fieldvalue" module, and when composing value renderers.

        See: https://docs.djangoproject.com/en/2.0/ref/templates/builtins/#std:templatetag-include
        """
        return self._value_renderer[mode]
        # return self.diagnostic_renderer(mode)

    def diagnostic_renderer(self, mode):
        """
        Return a diagnostic renderer for a specified mode.

        This is similar to the standard renderer, except that it allows for additional
        diagnostic information to be included to assist in debugging renderer call flows.
        """
        class _renderer(object):
            def __init__(self, field_renderer, mode):
                self._render_type = field_renderer._field_render_type
                self._value_mode  = field_renderer._field_value_mode
                self._mode        = mode
                self._renderer    = field_renderer._value_renderer[mode]
                return
            def render(self, context):
                try:
                    msg = (
                        "@@render(render_type=%s, value_mode=%s, mode=%s)"%
                        (self._render_type, self._value_mode, self._mode)
                        )
                    log.info(msg)
                    return "<!-- %s -->\n"%msg + self._renderer.render(context)
                except Exception as e:
                    log.error("Error in FieldRenderer.renderer: "+str(e))
                    log.info("\n".join(traceback.format_stack()))
                    return str(e)

        class _error_renderer(object):
            def __init__(self, field_renderer, mode, exc):
                self._render_type = field_renderer._field_render_type
                self._value_mode  = field_renderer._field_value_mode
                self._mode        = mode
                self._error       = str(exc)
                return
            def render(self, context):
                return (
                    "@@_error_renderer(render_type=%s, value_mode=%s, mode=%s, error=%s)"%
                    (self._render_type, self._value_mode, self._mode, self._error)
                    )

        try:
            r = _renderer(self, mode)
        except Exception as e:
            log.error("Error in FieldRenderer.renderer._renderer(): "+str(e))
            log.info("\n".join(traceback.format_stack()))
            r = _error_renderer(self, mode, e)
        return r

    def label(self):
        return self.renderer("label")

    def view(self):
        return self.renderer("view")

    def edit(self):
        return self.renderer("edit")

    def label_view(self):
        return self.renderer("label_view")

    def label_edit(self):
        return self.renderer("label_edit")

    def col_head(self):
        return self.renderer("col_head")

    def col_head_view(self):
        return self.renderer("col_head_view")

    def col_head_edit(self):
        return self.renderer("col_head_edit")

    def col_view(self):
        return self.renderer("col_view")

    def col_edit(self):
        return self.renderer("col_edit")

    def mode(self):
        class _renderer(object):
            def __init__(modeself):
                pass
            def render(modeself, context):
                try:
                    r = self._value_renderer[context['render_mode']].render(context)
                except Exception as e:
                    log.error("Exception in FieldRenderer.mode, _renderer.render: %s"%(str(e),))
                    formatted_traceback = traceback.format_stack()
                    log.info("".join(formatted_traceback))
                    response_parts = (
                        ["Exception in FieldRenderer.mode, _renderer.render"]+
                        [repr(e)]+
                        formatted_traceback
                        )
                    return "\n".join(response_parts)
                return r
        return _renderer()

# End.



"""
Tests for languaged-tagged text value rendering.
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2019, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import sys
import os
import unittest
import re

from annalist.resourcetypes import (
    file_extension, file_extension_for_content_type
    )

from annalist.views.fields.render_text_language import (
    get_text_language_renderer, 
    TextLanguageValueMapper
    )

from .field_rendering_support import FieldRendererTestSupport

class TextLanguageRenderingTest(FieldRendererTestSupport):

    def setUp(self):
        return

    def tearDown(self):
        return

    # Rendering test

    def test_RenderTextLanguageValue(self):

        def language_text_value(text, lang):
            return {'@value': text, '@language': lang} if lang else {'@value': text}

        def expect_render(text, lang):
            view_text = "%s (%s)"%(text, lang) if lang else text
            render_view = '''<span>%s</span>'''%(view_text)
            render_edit = (
                '''<input type="text" size="64" name="repeat_prefix_test_field" '''+
                       '''placeholder="(test placeholder)" '''+
                       '''value="%s" />'''
                )%view_text
            return {'view': render_view, 'edit': render_edit}

        test_values = (
            [ ("text-en",      "en",  "text-en (en)")
            , ("text-en",      "en",  "text-en    (en)")
            , ("text-nl",      "",    "text-nl")
            , ("text (en) nl", "",    "text (en) nl")
            ])
        test_value_context_renders = (
            [ (self._make_test_context(render_text), expect_render(text, lang))
              for text, lang, render_text in test_values
            ])
        renderer = get_text_language_renderer()

        for render_context, expect_render in test_value_context_renders:
            # print repr(render_context['field']['field_value'])
            self._check_value_renderer_results(
                renderer,
                context=render_context,
                expect_rendered_view=expect_render['view'],
                expect_rendered_edit=expect_render['edit'],
                collapse_whitespace=True
                )
        return

    def test_DecodeTextLanguageValue(self):
        test_decode_values = (
            { None:               { "@value": "" }
            , "text":             { "@value": "text" }
            , "text (en)":        { "@value": "text", "@language": "en" }
            , "text (en) more":   { "@value": "text (en) more" }
            , "":                 { "@value": "" }
            })
        for valtext, expect_valdata in test_decode_values.items():
            valdata = TextLanguageValueMapper.decode(valtext)
            self.assertEqual(
                valdata, expect_valdata, 
                "Value decode(%s) = %r, expect %r"%(valtext, valdata, expect_valdata)
                )
        return

# End.

if __name__ == "__main__":
    # import django
    # django.setup()  # Needed for template loader
    # Runtests in this module
    # runner = unittest.TextTestRunner(verbosity=2)
    # tests = unittest.TestSuite()
    # tests  = getSuite(select=sel)
    # if tests: runner.run(tests)
    unittest.main()



"""
Login and authentication views and related authentication setup logic
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2016, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os
import re
import traceback
import json
import markdown
import copy
import uuid
import urllib
from importlib import import_module

from django.core.urlresolvers import resolve, reverse
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.template import loader
from django.views import generic
from django.views.decorators.csrf import csrf_exempt

from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.models import User

from utils.http_errors      import error400values
from utils.py3porting       import urlparse, urljoin

from .                      import login_message
from .auth_django_client    import django_flow_from_user_id
from .auth_oidc_client      import oauth2_flow_from_provider_data
from .login_utils           import (
    HttpResponseRedirectWithQuery, 
    HttpResponseRedirectLogin
    )

PROVIDER_FILES = None

PROVIDER_DETAILS = None

settings = import_module(os.environ["DJANGO_SETTINGS_MODULE"])

def collect_provider_data():
    global PROVIDER_FILES, PROVIDER_DETAILS
    if PROVIDER_DETAILS is None:
        PROVIDER_DETAILS = {}
        PROVIDER_FILES   = {}
        clientsecrets_dirname = os.path.join(settings.CONFIG_BASE, "providers/")
        if os.path.isdir(clientsecrets_dirname):
            clientsecrets_files   = os.listdir(clientsecrets_dirname)
            for f in clientsecrets_files:
                if f.endswith(".json"):
                    provider_path = os.path.join(clientsecrets_dirname,f)
                    with open(provider_path, "r") as f:
                        provider_data = json.load(f)
                    provider_name = provider_data['web']['provider']
                    PROVIDER_FILES[provider_name]   = provider_path
                    PROVIDER_DETAILS[provider_name] = provider_data['web']
                    if 'provider_label' not in PROVIDER_DETAILS[provider_name]:
                        PROVIDER_DETAILS[provider_name]['provider_label'] = provider_name
                    log.debug("login_views: collect_provider_data %s"%(provider_name,))
                    # For debugging only: don't log details in running system...
                    # log.debug(json.dumps(
                    #     PROVIDER_DETAILS[provider_name], 
                    #     sort_keys=True,
                    #     indent=4,
                    #     separators=(',', ': ')
                    #     ))
    return

def confirm_authentication(view, 
        login_form_url=None, login_post_url=None, login_done_url=None, 
        user_profile_url=None, continuation_url=None, 
        help_path="annalist/views/help/"):
    """
    Return None if required authentication is present, otherwise
    a login redirection response to the supplied URI

    view.credential is set to credential that can be used to access resource

    Five URL parameters are passed in from the calling application:

    login_form_url      Page to gather information to initiate login process
    login_post_url      URL to which login information is posted
    login_done_url      URL retrieved with additional parameters when authentication
                        is complete (maybe failed). In the OAuth2 flow, this triggers
                        retrieval of user profile information.  Not used for local
                        authentication.
    user_profile_url    URL retrieved when user profile details have been set up.
    continuation_url    URL from which the login process was initiated.
    """
    if view.request.user.is_authenticated:
        return None
    if not login_form_url:
        return error400values(view, "No login form URI specified")
    if not login_done_url:
        return error400values(view, "No login completion URI specified")
    if not login_post_url:
        login_post_url = login_form_url
    if not continuation_url:
        continuation_url = view.request.path
    # Redirect to initiate login sequence 
    # See: https://docs.djangoproject.com/en/2.0/topics/http/sessions/
    view.request.session['login_form_url']   = login_form_url
    view.request.session['login_post_url']   = login_post_url
    view.request.session['login_done_url']   = login_done_url
    view.request.session['user_profile_url'] = user_profile_url
    view.request.session['continuation_url'] = continuation_url
    view.request.session['help_dir']         = os.path.join(settings.SITE_SRC_ROOT, help_path)
    userid = view.request.POST.get("userid", 
        view.request.GET.get("userid",
            view.request.session.get('login_recent_userid', "")
            )
        ) 
    query_params = (
        { "userid":           userid
        , "continuation_url": continuation_url
        })
    query_params.update(view.get_message_data())
    return HttpResponseRedirectWithQuery(login_form_url, query_params)

class LoginUserView(generic.View):
    """
    View class to present login form to gather user id and other login information.

    The login page solicits a user id and an identity provider

    The login page supports the following request parameters:

    continuation_url={uri}
    - a URL for a page that is displayed when the login process is complete.
    """

    def get(self, request):
        collect_provider_data()
        # @@TODO: check PROVIDER_FILES, report error if none here
        # Retrieve request parameters
        continuation_url  = request.GET.get("continuation_url", "/no-login-continuation/")
        # Check required values in session - if missing, restart sequence from original URI
        # This is intended to avoid problems if this view is invoked out of sequence
        login_post_url    = request.session.get("login_post_url",       None)
        login_done_url    = request.session.get("login_done_url",       None)
        user_profile_url  = request.session.get("user_profile_url",     None)
        help_dir          = request.session.get("help_dir",             None)
        recent_userid     = request.session.get("login_recent_userid",  "")
        if ( (login_post_url is None) or 
             (login_done_url is None) or 
             (user_profile_url is None) or 
             (help_dir is None) ):
            log.warning(
                "LoginUserView: missing details "+
                "login_post_url %s, login_done_url %s, user_profile_url %s, help_dir %s"%
                (login_post_url, login_done_url, user_profile_url, help_dir)
                )
            return HttpResponseRedirect(continuation_url)
        # Display login form
        default_provider = ""
        provider_tuples = (
            [ ( p.get('provider_order', 5),
                (k, p.get('provider_label', k), p.get('provider_image', None))
              ) for k, p in PROVIDER_DETAILS.items()
            ])
        provider_labels = [ p[1] for p in sorted(provider_tuples) ]
        for p in PROVIDER_DETAILS:
            if "default" in PROVIDER_DETAILS[p]:            
                default_provider = PROVIDER_DETAILS[p]["default"]
        logindata = (
            { "login_post_url":     login_post_url
            , "login_done_url":     login_done_url
            , "user_profile_url":   user_profile_url
            , "continuation_url":   continuation_url
            , "provider_keys":      list(PROVIDER_DETAILS)
            , "provider_labels":    provider_labels
            , "provider":           default_provider
            , "suppress_user":      True
            , "help_filename":      "login-help"
            , "userid":             request.GET.get("userid", recent_userid)
            , "info_head":          request.GET.get("info_head", None)
            , "info_message":       request.GET.get("info_message", None)
            , "error_head":         request.GET.get("error_head", None)
            , "error_message":      request.GET.get("error_message", None)
            })
        # Load help text if available
        if "help_filename" in logindata:
            help_filepath = help_dir + "%(help_filename)s.md"%(logindata)
            if os.path.isfile(help_filepath):
                with open(help_filepath, "r") as helpfile:
                    logindata["help_markdown"] = helpfile.read()
        if "help_markdown" in logindata:
            logindata["help_text"] = markdown.markdown(logindata["help_markdown"])
        # Render form & return control to browser
        template = loader.get_template("login.html")
        return HttpResponse(template.render(logindata, request=self.request))

class LoginPostView(generic.View):
    """
    View class to initiate an authentication flow, typically on  POST 
    of the login form.

    It saves the supplied user id in a session value, and redirects the user to the 
    identity provider, which in due course returns control to the application along 
    with a suitable authorization grant.

    The login form provides the following values:

    userid={string}
    - a user identifying string that will be associated with the external service
      login credentials.
    provider={string}
    - a string that identifies a provider selected to perform authentication
      of the indicated user.  This string is an index to PROVIDER_FILES,
      which in turn contains filenames for client secrets to use when accessing
      the indicated identity provider.
    login_done={uri}
    - a URI that is retrieved, with a suitable authorization grant as a parameter, 
      when appropriate permission has been confirmed by an authenticated user.
      Used to obtain user information following completion of authentication.
      Communicated via a hidden form value.
    user_profile_url={uri}
    - a URI that is retrieved, when user information has been obtained.  Expected use
      is to display user information, thenm continue tyo the page from which the
      login sequence was invoked.  Communicated via a hidden form value.
    continuation_url={uri}
    - URL of page from which logon sequence was invoked, and to which control is
      eventually returned.  Communicated via a hidden form value.
    """

    def post(self, request):
        log.info("LoginPostView.post")
        try:
            response = self.post_main(request)
        except Exception as e:
            # -- This should be redundant, but...
            log.error("Exception in LoginPostView.post (%r)"%(e))
            log.error("".join(traceback.format_stack()))
            # --
            log.exception(str(e))
            response = self.error(
                dict(self.error500values(),
                    message=str(e)+" - see server log for details"
                    )
                )
        log.info(
            "LoginPostView.post complete %d %s"%
            (response.status_code, response.reason_phrase)
            )
        return response

    def post_main(self, request):
        # Retrieve request parameters
        userid            = request.POST.get("userid",            "")
        provider          = request.POST.get("provider",          "No_provider")
        provider          = request.POST.get("login",             provider)
        login_done_url    = request.POST.get("login_done_url",    "/no_login_done_url_in_form/")
        user_profile_url  = request.POST.get("user_profile_url",  "/no_user_profile_url_in_form/")
        continuation_url  = request.POST.get("continuation_url",  "/no_continuation_url_in_form/")
        if request.POST.get("login", None):
            collect_provider_data()
            provider_data = PROVIDER_DETAILS[provider]
            provider_name = PROVIDER_FILES[provider]
            provider_mechanism = provider_data.get("mechanism", "OIDC")
            if userid and not re.match(r"\w+$", userid):
                return HttpResponseRedirectLogin(
                    request, 
                    login_message.USER_ID_SYNTAX%(userid)
                    )
            request.session['login_recent_userid']    = userid
            request.session['login_provider_data']    = provider_data
            request.session['login_continuation_url'] = continuation_url
            if provider_mechanism == "OIDC":
                # Create and initialize flow object
                log.debug("LoginPostView.post: SECURE_PROXY_SSL_HEADER %r"%(settings.SECURE_PROXY_SSL_HEADER,))
                log.debug("LoginPostView.post: scheme %s"%(request.scheme,))
                log.debug("LoginPostView.post: headers %r"%(request.META,))
                flow = oauth2_flow_from_provider_data(
                    provider_data,
                    redirect_uri=request.build_absolute_uri(login_done_url)
                    )
                request.session['oauth2_state']  = flow.step1_get_state_token()
                request.session['oauth2_userid'] = userid
                # Initiate OAuth2 dance
                # The response is handled by auth_oidc_client.OIDC_AuthDoneView
                return HttpResponseRedirect(flow.step1_get_authorize_url())
            if provider_mechanism == "django":
                flow = django_flow_from_user_id(
                    provider_data,
                    userid=userid,
                    auth_uri=reverse("LocalUserPasswordView"),
                    redirect_uri=request.build_absolute_uri(user_profile_url)
                    )
                # Initiate django authentication
                auth_uri = flow.step1_get_authorize_url()
                return HttpResponseRedirect(auth_uri)
            return HttpResponseRedirectLogin(
                request,
                login_message.UNRECOGNIZED_PROVIDER%(provider_mechanism, provider_name)
                )
        # Login cancelled: redirect to continuation
        # (which may just redisplay the login page)
        return HttpResponseRedirect(continuation_url)

class LogoutUserView(generic.View):
    """
    View class to handle logout
    """

    def get(self, request):
        recent_userid = request.session.get('login_recent_userid', "")
        logout(request)
        request.session['login_recent_userid'] = recent_userid
        continuation_url = request.GET.get("continuation_url", 
            urljoin(urlparse(request.path).path, "../")
            )
        return HttpResponseRedirect(continuation_url)

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
This module is used to cache per-collection information about entities of some
designated type.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

from annalist                       import layout
from annalist.exceptions            import Annalist_Error
from annalist.identifiers           import ANNAL, RDFS

from annalist.models.objectcache    import get_cache, remove_cache # , remove_matching_caches

#   ---------------------------------------------------------------------------
# 
#   Local helper functions
# 
#   ---------------------------------------------------------------------------

def make_cache_key(cache_type, entity_type_id, coll_id):
    return (cache_type, entity_type_id, coll_id)

def match_cache_key_unused_(cache_types, entity_cls):
    def match_fn(cachekey):
        return (cachekey[0] in cache_types) and (cachekey[1] == entity_cls._entitytypeid)
    return match_fn

#   ---------------------------------------------------------------------------
# 
#   Error class
# 
#   ---------------------------------------------------------------------------

class Cache_Error(Annalist_Error):
    """
    Class for errors raised by closure calculations.
    """
    def __init__(self, value=None, msg="Cache_error (collectionentityache)"):
        super(Cache_Error, self).__init__(value, msg)
        return

#   ---------------------------------------------------------------------------
# 
#   Entity-cache object class
# 
#   ---------------------------------------------------------------------------

class CollectionEntityCacheObject(object):
    """
    This class is an entity cache for a specified collection and entity type.

    NOTE: entities are instantiated with respect to a specified collection,
    but the collection objects are transient (regenerated for each request), so
    the cache stores the entity values but not the instantiated entities.

    Two kinds of information are cached:

    1.  entity cache: details of all entities that are visible in
        this class, indexed by entity id and entity URI
    2.  scope cache: lists of entity ids that are visible in different scopes: used 
        when returning entity enumerations (see method "get_all_entities").

    The scope cache is populated by calls to "get_all_entities".  When a entity is 
    added to or removed from the entity cache, lacking information about the scopes 
    where it is visible, the scope cache is cleared.

    Scope values currently include "user", "all", "site"; None => "coll".
    Apart from treating None as collection local scope, the logic in this class
    treats scope names as opaque identifiers.  The scope logic is embedded mainly
    in the Entity and EntityRoot class methods "_children".
    """

    _cache_types = {"entities_by_id", "entity_ids_by_uri", "entity_ids_by_scope"}

    def __init__(self, coll_id, entity_cls):
        """
        Initialize a cache object for a specified collection.

        coll_id         Collection id with which the entity cache is associated.
        """
        super(CollectionEntityCacheObject, self).__init__()
        self._coll_id             = coll_id
        self._entity_cls          = entity_cls
        self._type_id             = entity_cls._entitytypeid
        self._entities_by_id      = None
        self._entity_ids_by_uri   = None
        self._entity_ids_by_scope = None
        self._site_cache          = None
        return

    def _make_cache_key(self, cache_type):
        return make_cache_key(cache_type, self._type_id, self._coll_id)

    def _make_entity(self, coll, entity_id, entity_values):
        """
        Internal helper method to construct an entity given its Id and values.

        coll            is collection entity to which the new identity will belong
        entity_id       is the new entity id
        entity_values   is a dictionary containing:
                        ["parent_id"] is the id of the parent entity
                        ["data"] is a dictionary of values for the new entity

        Returns None if either Id or values evaluate as Boolean False (i.e. are None or empty).
        """
        entity = None
        if entity_id and entity_values:
            parent_id = entity_values["parent_id"]
            parent    = coll
            if coll.get_id() != parent_id:
                for parent in coll.get_alt_entities(altscope="all"):
                    if parent.get_id() == parent_id:
                        break
                else:
                    msg = (
                        "Saved parent id %s not found for entity %s/%s in collection %s"%
                        (parent_id, self._type_id, entity_id, coll.get_id())
                        )
                    log.error(msg)
                    raise ValueError(msg)
            entity = self._entity_cls._child_init(parent, entity_id)
            entity.set_values(entity_values["data"])
        return entity

    def _load_entity(self, coll, entity, entity_uri=None):
        """
        Internal helper method saves entity data to cache.
        This function does not actually read entity data.

        Returns True if new entity data is added, otherwise False.
        """
        entity_id     = entity.get_id()
        if not entity_uri:
            entity_uri    = entity.get_uri()
        entity_parent = entity.get_parent().get_id()
        entity_data   = entity.get_save_values()
        #@@TODO: remove this
        # def _load_entity_data(old_value):
        #     # Update other caches while _entities_by_id lock is acquired
        #     self._entity_ids_by_uri.set(entity_uri, entity_id)
        #     self._entity_ids_by_scope.flush()
        #     return {"parent_id": entity_parent, "data": entity_data}
        # add_entity, _ = self._entities_by_id.find(entity_id, _load_entity_data, None)
        #@@
        add_entity = False
        with self._entities_by_id.access(entity_id) as es:
            if entity_id not in es:
                # Update cache via context handler
                es[entity_id] = {"parent_id": entity_parent, "data": entity_data}
                # Update other caches while _entities_by_id lock is acquired
                self._entity_ids_by_uri.set(entity_uri, entity_id)
                self._entity_ids_by_scope.flush()
                add_entity = True
        return add_entity

    def _load_entities(self, coll):
        """
        Initialize cache of entities, if not already done.

        NOTE: site level entitites are cached separately by the collection cache 
        manager, and merged separately.  Hence "nosite" scope here.

        From entity.py:
            "nosite" - collection-level only: used for listing entities from just
                collections.  Used when cacheing data, where site data is assumed 
                to be invariant, hence no need to re-load.
        """
        scope_name = "nosite" if self._site_cache else "all"
        if self._entities_by_id is None:
            self._entities_by_id      = get_cache(self._make_cache_key("entities_by_id"))
            self._entity_ids_by_uri   = get_cache(self._make_cache_key("entity_ids_by_uri"))
            self._entity_ids_by_scope = get_cache(self._make_cache_key("entity_ids_by_scope"))
            for entity_id in coll._children(self._entity_cls, altscope=scope_name):
                t = self._entity_cls.load(coll, entity_id, altscope=scope_name)
                self._load_entity(coll, t)
        return

    def _drop_entity(self, coll, entity_id):
        """
        Drop entity from collection cache.

        Returns the entity removed, or None if not found.
        """
        entity_values = self._entities_by_id.get(entity_id, None)
        entity        = self._make_entity(coll, entity_id, entity_values)
        if entity:
            entity_uri = entity.get_uri()
            self._entities_by_id.pop(entity_id, None)
            self._entity_ids_by_uri.pop(entity_uri, None)
            self._entity_ids_by_scope.flush()
        return entity

    def set_site_cache(self, site_cache):
        self._site_cache = site_cache
        return

    def get_coll_id(self):
        return self._coll_id

    def set_entity(self, coll, entity):
        """
        Save a new or updated entity definition.
        """
        self._load_entities(coll)
        self._load_entity(coll, entity)
        return

    def remove_entity(self, coll, entity_id):
        """
        Remove entity from collection cache.

        Returns the entity removed, or None if not found.
        """
        self._load_entities(coll)       # @@TODO: is this needed?
        return self._drop_entity(coll, entity_id)

    def get_entity(self, coll, entity_id):
        """
        Retrieve the entity for a given entity id.

        Returns an entity for the supplied entity Id, or None 
        if not defined for the current collection.
        """
        self._load_entities(coll)
        entity_values = self._entities_by_id.get(entity_id, None)
        if entity_values:
            return self._make_entity(coll, entity_id, entity_values)
        # If not in collection cache, look for value in site cache:
        if self._site_cache:
            return self._site_cache.get_entity(coll.get_site_data(), entity_id)
        return None

    def get_entity_from_uri(self, coll, entity_uri):
        """
        Retrieve an entity for a given entity URI.

        Returns an entity for the specified collecion and entuty URI,
        or None if the entity URI does not exist
        """
        self._load_entities(coll)
        entity_id = self._entity_ids_by_uri.get(entity_uri, None)
        if entity_id:
            entity    = self.get_entity(coll, entity_id)
            return entity
        # If not in collection cache, look for value in site cache:
        if self._site_cache:
            return self._site_cache.get_entity_from_uri(
                coll.get_site_data(), entity_uri
                )
        return None

    def get_all_entity_ids(self, coll, altscope=None):
        """
        Returns an iterator over all entity ids currently defined for a collection, 
        which may be qualified by a specified scope.

        NOTE: this method returns only those entity ids for which a record has
        been saved to the collection data storage.
        """
        #@@TODO: remove this
        # def get_scope_ids(scope_entity_ids=[]):
        #     # Collect entity ids for named scope
        #     for entity_id in coll._children(self._entity_cls, altscope=altscope):
        #         if entity_id != layout.INITIAL_VALUES_ID:
        #             scope_entity_ids.append(entity_id)
        #     return scope_entity_ids
        # self._load_entities(coll)
        # scope_name = altscope or "coll"     # 'None' designates collection-local scope
        # _, entity_ids = self._entity_ids_by_scope.find(scope_name, get_scope_ids, [])
        #@@
        self._load_entities(coll)
        scope_name = altscope or "coll"     # 'None' designates collection-local scope
        scope_entity_ids = []
        with self._entity_ids_by_scope.access(scope_name) as eids:
            if scope_name in eids:
                scope_entity_ids = eids[scope_name]
            else:
                # Collect entity ids for named scope
                for entity_id in coll._children(self._entity_cls, altscope=altscope):
                    if entity_id != layout.INITIAL_VALUES_ID:
                        scope_entity_ids.append(entity_id)
                # Update cache via context manager
                eids[scope_name] = scope_entity_ids
        return scope_entity_ids

    def get_all_entities(self, coll, altscope=None):
        """
        Returns a generator of all entities currently defined for a collection, which
        may be qualified by a specified scope.

        NOTE: this method returns only those records that have actually been saved to
        the collection data storage.
        """
        scope_entity_ids = self.get_all_entity_ids(coll, altscope=altscope)
        for entity_id in scope_entity_ids:
            t = self.get_entity(coll, entity_id)
            if t:
                yield t
        return

    def remove_cache(self):
        """
        Close down and release all entity cache data
        """
        if self._entities_by_id:
            remove_cache(self._entities_by_id.cache_key())
            self._entities_by_id = None
        if self._entity_ids_by_uri:
            remove_cache(self._entity_ids_by_uri.cache_key())
            self._entity_ids_by_uri = None
        if self._entity_ids_by_scope:
            remove_cache(self._entity_ids_by_scope.cache_key())
            self._entity_ids_by_scope = None
        return

#   ---------------------------------------------------------------------------
# 
#   Collection entity-cache class
# 
#   ---------------------------------------------------------------------------

coll_cache_by_type_id_coll_id = {}

class CollectionEntityCache(object):
    """
    This class manages multiple-collection cache objects
    """
    def __init__(self, cache_cls, entity_cls):
        """
        Initializes a value cache with no per-collection data.

        cache_cls   is a class object for the collaction cache objects to be used.
                    The constructor is called with collection id and entity class
                    as parameters (see method `_get_cache`).
        entity_cls  is a class object for the type of entity to be cached.
        """
        super(CollectionEntityCache, self).__init__()
        self._cache_cls  = cache_cls
        self._entity_cls = entity_cls
        self._type_id    = entity_cls._entitytypeid
        coll_cache_by_type_id_coll_id[self._type_id] = {}
        return

    # Generic collection cache alllocation and access methods

    def _get_site_cache(self):
        """
        Local helper returns a cache object for the site-wide entities
        """
        if layout.SITEDATA_ID not in coll_cache_by_type_id_coll_id[self._type_id]:
            # log.info(
            #     "CollectionEntityCache: creating %s cache for collection %s"%
            #     (self._type_id, layout.SITEDATA_ID)
            #     )
            # Create and save new cache object
            site_cache = self._cache_cls(layout.SITEDATA_ID, self._entity_cls)
            coll_cache_by_type_id_coll_id[self._type_id][layout.SITEDATA_ID] = site_cache
        return coll_cache_by_type_id_coll_id[self._type_id][layout.SITEDATA_ID]

    def _get_cache(self, coll):
        """
        Local helper returns a cache object for a specified collection.

        Creates a new cache object if needed.

        coll        is a collection object for which a cache object is obtained
        """
        coll_id = coll.get_id()
        # log.info(
        #     "CollectionEntityCache: get %s cache for collection %s"%
        #     (self._type_id, coll_id)
        #     )
        if coll_id not in coll_cache_by_type_id_coll_id[self._type_id]:
            # log.debug(
            #     "CollectionEntityCache: creating %s cache for collection %s"%
            #     (self._type_id, coll_id)
            #     )
            # Create and save new cache object
            coll_cache = self._cache_cls(coll_id, self._entity_cls)
            coll_cache.set_site_cache(self._get_site_cache())
            coll_cache_by_type_id_coll_id[self._type_id][coll_id] = coll_cache
        return coll_cache_by_type_id_coll_id[self._type_id][coll_id]

    def flush_cache(self, coll):
        """
        Remove all cached data for a specified collection.

        Returns True if the cache object was defined, otherwise False.

        coll        is a collection object for which a cache is removed.
        """
        coll_id = coll.get_id()
        cache   = coll_cache_by_type_id_coll_id[self._type_id].pop(coll_id, None)
        if cache:
            cache.remove_cache()
            # log.info(
            #     "CollectionEntityCache: flushed %s cache for collection %s"%
            #     (self._type_id, coll_id)
            #     )
            return True
        return False

    def flush_all(self):
        """
        Remove all cached data for all collections.
        """
        # remove_cache_types = CollectionEntityCacheObject._cache_types
        # find_matching_caches(
        #     match_cache_key(remove_cache_types, entity_cls),
        #     lambda cache: cache.close()
        #     )
        caches = coll_cache_by_type_id_coll_id[self._type_id]
        coll_cache_by_type_id_coll_id[self._type_id] = {}
        # log.info(
        #     "CollectionEntityCache: flushing %s cache for collections %r"%
        #     (self._type_id, caches.keys())
        #     )
        for coll_id in caches:
            caches[coll_id].remove_cache()
        return

    # Collection cache allocation and access methods

    def set_entity(self, coll, entity):
        """
        Save a new or updated type definition
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.set_entity(coll, entity)

    def remove_entity(self, coll, entity_id):
        """
        Remove entity from collection cache.

        Returns the entity removed if found, or None if not defined.
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.remove_entity(coll, entity_id)

    def get_entity(self, coll, entity_id):
        """
        Retrieve an entity for a given entity id.

        Returns an entity object for the specified collection and entity id.
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.get_entity(coll, entity_id)

    def get_entity_from_uri(self, coll, entity_uri):
        """
        Retrieve en entity for a given collection and entity URI.

        Returns an entity object for the specified collection and entity URI.
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.get_entity_from_uri(coll, entity_uri)

    def get_all_entity_ids(self, coll, altscope=None):
        """
        Returns all entities currently available for a collection in the indicated scope.
        Default scope is entities defined directly in the indicated collection.
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.get_all_entity_ids(coll, altscope=altscope)

    def get_all_entities(self, coll, altscope=None):
        """
        Returns all entities currently available for a collection in the indicated scope.
        Default scope is entities defined directly in the indicated collection.
        """
        entity_cache = self._get_cache(coll)
        return entity_cache.get_all_entities(coll, altscope=altscope)

# End.



"""
OAuth2 / OpenID Connect authentication related view handler and
supporting utilities.

NOTE: for Google provider, set up via 
https://console.developers.google.com/apis/dashboard
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2016, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import sys
import os
import json
import re
import traceback
import logging
log = logging.getLogger(__name__)

from requests_oauthlib import OAuth2Session

from django.core.urlresolvers import resolve, reverse
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.views import generic
from django.views.decorators.csrf import csrf_exempt

from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.models import User

from utils.http_errors import error400values

from .                  import login_message
from .login_utils       import (
    HttpResponseRedirectWithQuery, 
    HttpResponseRedirectLogin
    )

SCOPE_DEFAULT = ["openid", "profile", "email"]

#   ---------------------------------------------------------------------------
#
#   Factory function
#
#   ---------------------------------------------------------------------------

def oauth2_flow_from_provider_data(provider_data, redirect_uri=None, state=None):
    """
    Create an OpenId connect Oauth2 flow object from a supplied provider details file.

    provider_data
            dictionary containing provider details (including oauth2 client secrets).
    redirect_uri
            URI to which control is transferred when the OAuth2 authetication dance 
            is completed.  If specified, overrides value from provider-file.
    """
    return oauth2_flow(provider_data, redirect_uri=redirect_uri, state=state)

#   ---------------------------------------------------------------------------
#
#   Oauth2 flow object (based loosely on oauth2client flow object)
#
#   ---------------------------------------------------------------------------

class oauth2_flow(object):
    """
    Choreographs the oauth2 dance used to obtain a user authentication credential.
    """

    def __init__(self, provider_data, scope=None, state=None, redirect_uri=None):
        """
        Initialize a flow object with supplied provider data (provided as a dictionary)
        """
        self._provider_data = provider_data
        self._scope         = scope or provider_data.get("scope", SCOPE_DEFAULT)
        assert redirect_uri, "Redirect URI not specified"
        self._redirect_uri  = redirect_uri
        session             = OAuth2Session(
                                client_id=provider_data["client_id"],
                                scope=self._scope,
                                state=state,
                                redirect_uri=self._redirect_uri
                                )
        auth_uri, state     = session.authorization_url(provider_data["auth_uri"])
        self._session       = session
        self._auth_uri      = auth_uri
        self._state         = state
        # log.debug("oauth2_flow: provider_data %r"%(self._provider_data,))
        log.debug("oauth2_flow: scope         %r"%(self._scope,))
        log.debug("oauth2_flow: redirect_uri  %r"%(self._redirect_uri,))
        log.debug("oauth2_flow: session       %r"%(self._session,))
        log.debug("oauth2_flow: auth_uri      %r"%(self._auth_uri,))
        log.debug("oauth2_flow: state         %r"%(self._state,))
        return

    def step1_get_state_token(self):
        return self._state

    def step1_get_authorize_url(self):
        log.info("step1_get_authorize_url: auth_uri %r", self._auth_uri)
        return self._auth_uri

    def step2_exchange(self, request):
        """
        Using a credentials provided in the supplied redirect request value,
        requests an access token for user authentication.
        """
        auth_resp     = request.build_absolute_uri()
        token_uri     = self._provider_data['token_uri']
        client_secret = self._provider_data['client_secret']
        log.debug("step2_exchange: token_uri     %r", token_uri)
        log.debug("step2_exchange: auth_resp     %r", auth_resp)
        # For debugging onlky.  Don't log this in a running system!
        # log.debug("step2_exchange: client_secret %r", client_secret)
        try:
            token = self._session.fetch_token(token_uri, 
                client_secret=client_secret,
                authorization_response=auth_resp,
                timeout=5
                )
        except Exception as e:
            log.error("Failed to fetch token: %s"%(e,))
            # log.info(json.dumps(
            #     self._provider_data,
            #     sort_keys=True,
            #     indent=4,
            #     separators=(',', ': ')
            #     ))
            raise
        return token

    def step3_get_profile(self, token):
        """
        Uses saved credentials from `step2_exchange` to access the user profile,
        which is returned as a dictionary.  The content is determined by the identity
        provider service, but is expected to contain at least:

            { "verified_email": true,
              "email":          "...",
              "given_name":     "...",
              "family_name":    "...",
            }
        """
        r = self._session.get(self._provider_data["profile_uri"])
        profile = json.loads(r.content)
        return profile

class OIDC_AuthDoneView(generic.View):
    """
    View class used to complete login process with authorization grant provided by
    OAuth2 authorization server.

    The calling application must set up the URL routing for this handler to be invoked.
    """

    def get(self, request):
        # Look for authorization grant
        provider_data  = request.session['login_provider_data']
        state          = request.session['oauth2_state']
        userid         = request.session['oauth2_userid']
        # session value "login_done_url" is set by login_views.confirm_authentication
        if 'login_done_url' not in request.session:
            return HttpResponseRedirectLogin(request, login_message.SESSION_INTERRUPTED)
        login_done_url = request.build_absolute_uri(request.session['login_done_url'])
        provider       = provider_data['provider']
        flow = oauth2_flow_from_provider_data(
                provider_data, 
                redirect_uri=login_done_url, 
                state=state
                )
        # Get authenticated user details
        try:
            credential = flow.step2_exchange(request)
            profile    = flow.step3_get_profile(credential)
            log.debug("auth_oidc_client: userid %s, profile %r"%(userid, profile))
            authuser = authenticate(
                username=userid, profile=profile
                )
        except Exception as e:
            log.error("Exception %r"%(e,))
            # For debugging only: don't log in running system
            # log.error("provider_data %r"%(provider_data,))
            ex_type, ex, tb = sys.exc_info()
            log.error("".join(traceback.format_exception(ex_type, ex, tb)))
            # log.error("".join(traceback.format_stack()))
            return HttpResponseRedirectLogin(request, str(e))
        # Check authenticated details for user id match any previous values.
        #
        # The user id is entered by the user on the login form, and is used as a key to
        # access authenticated user details in the Django user database.  The user id 
        # itself is not checked by the Oauth2 login flow, other than for checking that
        # it contains only work characters
        #
        # Instead, we trust that the associated email address has been confirmed by the 
        # OAuth2 provider, and don't allow login where the email adress differs from any 
        # currently saved email address for the user id used..  This aims to  prevent a 
        # new set of OAuth2 credentials being used for a previously created Django user id.
        #
        if not authuser:
            return HttpResponseRedirectLogin(request, 
                login_message.USER_NOT_AUTHENTICATED%(userid, provider)
                )
        if not userid:
            # Get generated username
            userid = authuser.username
        if not re.match(r"\w+$", userid):
            return HttpResponseRedirectLogin(
                request, 
                login_message.USER_ID_SYNTAX%(userid)
                )
        if not authuser.email:
            return HttpResponseRedirectLogin(request, 
                login_message.USER_NO_EMAIL%(userid)
                )
        try:
            olduser = User.objects.get(username=userid)
        except User.DoesNotExist:
            olduser = None
        if olduser:
            if authuser.email != olduser.email:
                return HttpResponseRedirectLogin(request, 
                    login_message.USER_WRONG_EMAIL%(userid, authuser.email, olduser.email)
                    )
        # Complete the login and save details
        authuser.save()
        login(request, authuser)
        request.session['login_recent_userid'] = userid
        log.info("OIDC_AuthDoneView: user.username:   "+authuser.username)
        log.info("OIDC_AuthDoneView: user.first_name: "+authuser.first_name)
        log.info("OIDC_AuthDoneView: user.last_name:  "+authuser.last_name)
        log.info("OIDC_AuthDoneView: user.email:      "+authuser.email)
        return HttpResponseRedirectLogin(request)

# End.



"""
Development settings

Data is kept within the project directory
(initialize as required, e.g. by copying initial testdata).
Service configuration is kept under personal home directory to 
protect secret keys, etc.
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

from .common import *

ANNALIST_VERSION_MSG = "Annalist version %s (development configuration)"%(ANNALIST_VERSION)

SETTINGS_MODULE   = __name__
SITE_DIR_NAME     = "annalist_site"
BASE_DATA_DIR     = SITE_SRC_ROOT+"/devel"
BASE_SITE_DIR     = os.path.join(BASE_DATA_DIR, SITE_DIR_NAME)
CONFIG_BASE       = os.path.join(os.path.expanduser("~"), ".annalist/")
STATIC_ROOT       = os.path.join(BASE_SITE_DIR, 'static')

BASE_LOG_DIR      = BASE_SITE_DIR+"/"
ANNALIST_LOG_PATH = "None (output to console)"
ACCESS_LOG_PATH   = "None"
ERROR_LOG_PATH    = "None"

DATABASE_PATH     = os.path.join(BASE_SITE_DIR, 'db.sqlite3')
DATABASES         = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': DATABASE_PATH,
    }
}

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

# import logging
# log = logging.getLogger(__name__)
# log.info("Annalist version %s (development configuration)"%(ANNALIST_VERSION))
# log.info(ANNALIST_VERSION_MSG)
# log.info("SETTINGS_MODULE:  "+SETTINGS_MODULE)
# log.info("BASE_DATA_DIR:    "+BASE_DATA_DIR)
# log.info("CONFIG_BASE:      "+CONFIG_BASE)
# log.info("DJANGO_ROOT:      "+DJANGO_ROOT)
# log.info("SITE_CONFIG_DIR:  "+SITE_CONFIG_DIR)
# log.info("SITE_SRC_ROOT:    "+SITE_SRC_ROOT)
# log.info("STATICFILES_DIRS: "+repr(STATICFILES_DIRS))
# log.info("DB PATH:          "+DATABASES['default']['NAME'])

# End.



"""
Test module for annalist-manager information commands
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import sys
import os

from utils.py3porting           import StringIO
from utils.StdoutContext        import SwitchStdout, SwitchStderr

import annalist

from annalist_manager.tests     import get_source_root
from annalist_manager.tests     import test_annalist_base
from annalist_manager.am_main   import runCommand

#   -----------------------------------------------------------------------------
#
#   Tests
#
#   -----------------------------------------------------------------------------

class AnnalistManagerInfoTest(test_annalist_base.AnnalistManagerTestBase):

    @classmethod
    def setUpTestData(cls):
        cls.setup_annalist_manager_test()
        return

    def setUp(self):
        # self.userhome    = os.path.os.path.expanduser("~")
        # self.userconfig  = os.path.os.path.expanduser("~/.annalist")
        # self.src_root    = get_source_root()
        # self.testhome    = os.path.join(self.src_root, "sampledata/data")
        # self.settingsdir = os.path.join(self.src_root, "annalist_site/settings")
        return

    def tearDown(self):
        return

    #   -----------------------------------------------------------------------------
    #   Tests
    #   -----------------------------------------------------------------------------

    def test_Version(self):
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "version"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], annalist.__version__)
        return

    def test_Help(self):
        stdoutbuf  = StringIO()
        with SwitchStderr(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "help"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[1], "Commands:")
        self.assertEqual(stdoutlines[3], "  annalist-manager help [command]")
        self.assertEqual(stdoutlines[4], "  annalist-manager runtests [testlabel]")
        self.assertEqual(stdoutlines[5], "  annalist-manager initialize [ CONFIG ]")
        stdoutbuf.seek(0)
        with SwitchStderr(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "help", "init"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[1], "  annalist-manager initialize [ CONFIG ]")
        self.assertEqual(stdoutlines[3], "Initializes the installed software for an indicated configuration.")
        self.assertEqual(stdoutlines[4], "Mainly, this involves creating the internal database used to manage users, etc.")
        self.assertEqual(stdoutlines[6], "Annalist can be run in a number of configurations, notably")
        self.assertEqual(stdoutlines[7], "'development', 'personal' and 'shared'.")
        return

    def test_SiteDirectory(self):
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "sitedirectory"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.userhome+"/annalist_site")
        stdoutbuf.seek(0)
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "sitedirectory", "--config=runtests"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.sitehome)
        return

    def test_SettingsModule(self):
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "settingsmodule"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], "annalist_site.settings.personal")
        stdoutbuf.seek(0)
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "settingsmodule", "--config=runtests"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], "annalist_site.settings.runtests")
        return

    def test_SettingsDirectory(self):
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "settingsdirectory"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.settingsdir)
        stdoutbuf.seek(0)
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "settingsdirectory", "--config=runtests"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.settingsdir)
        return

    def test_SettingsFile(self):
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "settingsfile"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.settingsdir+"/personal")
        stdoutbuf.seek(0)
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "settingsfile", "--config=runtests"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.settingsdir+"/runtests")
        return

    def test_ServerLog(self):
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "serverlog"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.userhome+"/annalist_site/annalist.log")
        stdoutbuf.seek(0)
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "serverlog", "--config=runtests"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.src_root+"/annalist.log")
        return

    def test_AccessLog(self):
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "accesslog"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.userhome+"/annalist_site/annalist-wsgi-access.log")
        stdoutbuf.seek(0)
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "accesslog", "--config=runtests"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.src_root+"/annalist-wsgi-access.log")
        return

    def test_ErrorLog(self):
        stdoutbuf  = StringIO()
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "errorlog"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.userhome+"/annalist_site/annalist-wsgi-error.log")
        stdoutbuf.seek(0)
        with SwitchStdout(stdoutbuf):
            runCommand(self.userhome, self.userconfig, 
                ["annalist-manager", "errorlog", "--config=runtests"]
                )
        stdoutbuf.seek(0)
        stdoutlines = stdoutbuf.read().split("\n")
        self.assertEqual(stdoutlines[0], self.src_root+"/annalist-wsgi-error.log")
        return

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Annalist class for processing a RepeatValuesMap in an annalist view value mapping table.

A RepeatValuesMap is used to render repeated groups of fields for each of 
a list of values.

When the mapping function `map_entity_to_context` is called, the supplied `entityvals` 
is expected to be an iterator (list, tuple, etc.) of entities or dictionary values 
to be processed for display.

When decoding values from a form, different logic is required to extract a
repeating structure from the flat namespace used for form data.  See method 
`map_form_to_entity`, along with `FieldListValueMap.map_form_to_entity_repeated_item` 
for more details. 
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

# import json

from annalist.identifiers               import RDFS, ANNAL

from annalist.views.form_utils.fieldvaluemap    import FieldValueMap
from annalist.views.fields.bound_field          import bound_field

class RepeatValuesMap(FieldValueMap):
    """
    Define an entry in an entity value mapping table corresponding to a
    group of fields that is repeated for multiple values in an entity.  

    c           request context field name for the field value mapping entry
    f           is a `FieldDescription` value describing the repeated data.
    fieldlist   a `FieldListValueMap` object describing a set of fields to be 
                displayed for each repeated value.
    """

    def __init__(self, c=None, f=None, fieldlist=None):
        super(RepeatValuesMap, self).__init__(c=c, f=f)
        self.fieldlist = fieldlist
        return

    def __repr__(self):
        return (
            super(RepeatValuesMap, self).__repr__()+
            "RepeatValuesMap.fieldlist: %r\n"%(self.fieldlist)
            )

    def map_form_to_entity(self, formvals, entityvals):
        # log.info(repr(formvals))
        prefix_template  = self.i+"__%d__"
        prefix_n         = 0
        repeat_vals      = []
        field_key        = self.f.get_field_value_key(entityvals)
        previous_vals    = entityvals.get(field_key,[])
        while True:
            #@@ 
            #   The following logic works, but sometimes yields unexpected results:
            #   Specifically, when the original data has more fields than the form,
            #   the additional old fields were copied over into the result.
            #
            #   For now, we live with the restriction that fields within repeated 
            #   fields cannot propagate subproperty values used; i.e. when editing, 
            #   subproperties used in the data are replaced by the superproperty from
            #   the repeated field definition.  In practice, this may not be a problem,
            #   as the cases of repeated fields with subproperties are generally associated
            #   with special JSON-LD keys like '@id' or '@value'
            #
            # Extract previous values in same position to be updated
            # This ad-hocery is used to try and preserve property URIs used 
            # within the list, so that subproperties (where used) are preserved.
            # if len(previous_vals) > prefix_n:
            #     vals     = previous_vals[prefix_n]
            # else:
            #     vals     = {}
            #@@
            vals         = {}
            prefix       = prefix_template%prefix_n
            updated_vals = self.fieldlist.map_form_to_entity_repeated_item(
                formvals, vals, prefix
                )
            if not updated_vals:
                break
            repeat_vals.append(updated_vals)
            prefix_n += 1
        entityvals[field_key] = repeat_vals
        return entityvals

    def get_structure_description(self):
        """
        Helper function returns structure description information
        """
        return (
            { 'field_type':     "RepeatValuesMap"
            , 'field_descr':    self.f
            , 'entity_field':   self.e
            , 'form_field':     self.i
            })

# End.
